from typing import Final

from pyteal.types import TealType, require_type
from pyteal.ir import Op
from pyteal.ast.expr import Expr
from pyteal.ast.maybe import MaybeValue


class AssetHolding:
    @classmethod
    def balance(cls, account: Expr, asset: Expr) -> MaybeValue:
        """Get the amount of an asset held by an account.

        Args:
            account: An index into Txn.Accounts that corresponds to the account to check,
                must be evaluated to uint64 (or, since v4, an account address that appears in
                Txn.Accounts or is Txn.Sender, must be evaluated to bytes).
            asset: The ID of the asset to get, must be evaluated to uint64 (or, since v4,
                a Txn.assets offset).
        """
        require_type(account, TealType.anytype)
        require_type(asset, TealType.uint64)
        return MaybeValue(
            Op.asset_holding_get,
            TealType.uint64,
            immediate_args=["AssetBalance"],
            args=[account, asset],
        )

    @classmethod
    def frozen(cls, account: Expr, asset: Expr) -> MaybeValue:
        """Check if an asset is frozen for an account.

        A value of 1 indicates frozen and 0 indicates not frozen.

        Args:
            account: An index into Txn.Accounts that corresponds to the account to check,
                must be evaluated to uint64 (or, since v4, an account address that appears in
                Txn.Accounts or is Txn.Sender, must be evaluated to bytes).
            asset: The ID of the asset to get, must be evaluated to uint64 (or, since v4,
                a Txn.assets offset).
        """
        require_type(account, TealType.anytype)
        require_type(asset, TealType.uint64)
        return MaybeValue(
            Op.asset_holding_get,
            TealType.uint64,
            immediate_args=["AssetFrozen"],
            args=[account, asset],
        )


AssetHolding.__module__ = "pyteal"


class AssetHoldingObject:
    """Represents information about an account's holding of an asset"""

    def __init__(self, asset: Expr, account: Expr) -> None:
        """Create a new AssetParamObject for the given asset.

        Args:
            asset: An identifier for the asset. It must be an index into Txn.ForeignAssets that
                corresponds to the asset to check, or since v4, an asset ID that appears in
                Txn.ForeignAssets. In either case, it must evaluate to uint64.
            account: An identifier for the account. It must be an index into Txn.Accounts that
                corresponds to the account to check (in which case it must evaluate to uint64), or
                since v4, an account address that appears in Txn.Accounts or is Txn.Sender (in which
                case it must evaluate to bytes).
        """
        require_type(asset, TealType.uint64)
        self._asset: Final = asset
        require_type(account, TealType.anytype)
        self._account: Final = account

    def balance(self) -> MaybeValue:
        """Get the amount of the asset held by the account."""
        return AssetHolding.balance(self._account, self._asset)

    def frozen(self) -> MaybeValue:
        """Check if the asset is frozen for the account.

        A value of 1 indicates frozen and 0 indicates not frozen.
        """
        return AssetHolding.frozen(self._account, self._asset)


AssetHoldingObject.__module__ = "pyteal"


class AssetParam:
    @classmethod
    def total(cls, asset: Expr) -> MaybeValue:
        """Get the total number of units of an asset.

        Args:
            asset: An index into Txn.assets that corresponds to the asset to check,
                must be evaluated to uint64 (or since v4, an asset ID that appears in
                Txn.assets).
        """
        require_type(asset, TealType.uint64)
        return MaybeValue(
            Op.asset_params_get,
            TealType.uint64,
            immediate_args=["AssetTotal"],
            args=[asset],
        )

    @classmethod
    def decimals(cls, asset: Expr) -> MaybeValue:
        """Get the number of decimals for an asset.

        Args:
            asset: An index into Txn.assets that corresponds to the asset to check,
                must be evaluated to uint64 (or since v4, an asset ID that appears in
                Txn.assets).
        """
        require_type(asset, TealType.uint64)
        return MaybeValue(
            Op.asset_params_get,
            TealType.uint64,
            immediate_args=["AssetDecimals"],
            args=[asset],
        )

    @classmethod
    def defaultFrozen(cls, asset: Expr) -> MaybeValue:
        """Check if an asset is frozen by default.

        Args:
            asset: An index into Txn.assets that corresponds to the asset to check,
                must be evaluated to uint64 (or since v4, an asset ID that appears in
                Txn.assets).
        """
        require_type(asset, TealType.uint64)
        return MaybeValue(
            Op.asset_params_get,
            TealType.uint64,
            immediate_args=["AssetDefaultFrozen"],
            args=[asset],
        )

    @classmethod
    def unitName(cls, asset: Expr) -> MaybeValue:
        """Get the unit name of an asset.

        Args:
            asset: An index into Txn.assets that corresponds to the asset to check,
                must be evaluated to uint64 (or since v4, an asset ID that appears in
                Txn.assets).
        """
        require_type(asset, TealType.uint64)
        return MaybeValue(
            Op.asset_params_get,
            TealType.bytes,
            immediate_args=["AssetUnitName"],
            args=[asset],
        )

    @classmethod
    def name(cls, asset: Expr) -> MaybeValue:
        """Get the name of an asset.

        Args:
            asset: An index into Txn.assets that corresponds to the asset to check,
                must be evaluated to uint64 (or since v4, an asset ID that appears in
                Txn.assets).
        """
        require_type(asset, TealType.uint64)
        return MaybeValue(
            Op.asset_params_get,
            TealType.bytes,
            immediate_args=["AssetName"],
            args=[asset],
        )

    @classmethod
    def url(cls, asset: Expr) -> MaybeValue:
        """Get the URL of an asset.

        Args:
            asset: An index into Txn.assets that corresponds to the asset to check,
                must be evaluated to uint64 (or since v4, an asset ID that appears in
                Txn.assets).
        """
        require_type(asset, TealType.uint64)
        return MaybeValue(
            Op.asset_params_get,
            TealType.bytes,
            immediate_args=["AssetURL"],
            args=[asset],
        )

    @classmethod
    def metadataHash(cls, asset: Expr) -> MaybeValue:
        """Get the arbitrary commitment for an asset.

        If set, this will be 32 bytes long.

        Args:
            asset: An index into Txn.assets that corresponds to the asset to check,
                must be evaluated to uint64 (or since v4, an asset ID that appears in
                Txn.assets).
        """
        require_type(asset, TealType.uint64)
        return MaybeValue(
            Op.asset_params_get,
            TealType.bytes,
            immediate_args=["AssetMetadataHash"],
            args=[asset],
        )

    @classmethod
    def manager(cls, asset: Expr) -> MaybeValue:
        """Get the manager address for an asset.

        Args:
            asset: An index into Txn.assets that corresponds to the asset to check,
                must be evaluated to uint64 (or since v4, an asset ID that appears in
                Txn.assets).
        """
        require_type(asset, TealType.uint64)
        return MaybeValue(
            Op.asset_params_get,
            TealType.bytes,
            immediate_args=["AssetManager"],
            args=[asset],
        )

    @classmethod
    def reserve(cls, asset: Expr) -> MaybeValue:
        """Get the reserve address for an asset.

        Args:
            asset: An index into Txn.assets that corresponds to the asset to check,
                must be evaluated to uint64 (or since v4, an asset ID that appears in
                Txn.assets).
        """
        require_type(asset, TealType.uint64)
        return MaybeValue(
            Op.asset_params_get,
            TealType.bytes,
            immediate_args=["AssetReserve"],
            args=[asset],
        )

    @classmethod
    def freeze(cls, asset: Expr) -> MaybeValue:
        """Get the freeze address for an asset.

        Args:
            asset: An index into Txn.assets that corresponds to the asset to check,
                must be evaluated to uint64 (or since v4, an asset ID that appears in
                Txn.assets).
        """
        require_type(asset, TealType.uint64)
        return MaybeValue(
            Op.asset_params_get,
            TealType.bytes,
            immediate_args=["AssetFreeze"],
            args=[asset],
        )

    @classmethod
    def clawback(cls, asset: Expr) -> MaybeValue:
        """Get the clawback address for an asset.

        Args:
            asset: An index into Txn.assets that corresponds to the asset to check,
                must be evaluated to uint64 (or since v4, an asset ID that appears in
                Txn.assets).
        """
        require_type(asset, TealType.uint64)
        return MaybeValue(
            Op.asset_params_get,
            TealType.bytes,
            immediate_args=["AssetClawback"],
            args=[asset],
        )

    @classmethod
    def creator(cls, asset: Expr) -> MaybeValue:
        """Get the creator address for an asset.

        Args:
            asset: An index into Txn.assets that corresponds to the asset to check. Must
                evaluate to uint64.
        """
        require_type(asset, TealType.uint64)
        return MaybeValue(
            Op.asset_params_get,
            TealType.bytes,
            immediate_args=["AssetCreator"],
            args=[asset],
        )


AssetParam.__module__ = "pyteal"


class AssetParamObject:
    """Represents information about an asset's parameters"""

    def __init__(self, asset: Expr) -> None:
        """Create a new AssetParamObject for the given asset.

        Args:
            asset: An identifier for the asset. It must be an index into Txn.ForeignAssets that
                corresponds to the asset to check, or since v4, an asset ID that appears in
                Txn.ForeignAssets. In either case, it must evaluate to uint64.
        """
        require_type(asset, TealType.uint64)
        self._asset: Final = asset

    def total(self) -> MaybeValue:
        """Get the total number of units of the asset."""
        return AssetParam.total(self._asset)

    def decimals(self) -> MaybeValue:
        """Get the number of decimals for the asset."""
        return AssetParam.decimals(self._asset)

    def default_frozen(self) -> MaybeValue:
        """Check if the asset is frozen by default."""
        return AssetParam.defaultFrozen(self._asset)

    def unit_name(self) -> MaybeValue:
        """Get the unit name of the asset."""
        return AssetParam.unitName(self._asset)

    def name(self) -> MaybeValue:
        """Get the name of the asset."""
        return AssetParam.name(self._asset)

    def url(self) -> MaybeValue:
        """Get the URL of the asset."""
        return AssetParam.url(self._asset)

    def metadata_hash(self) -> MaybeValue:
        """Get the arbitrary commitment for the asset.

        If set, this will be 32 bytes long."""
        return AssetParam.metadataHash(self._asset)

    def manager_address(self) -> MaybeValue:
        """Get the manager address for the asset."""
        return AssetParam.manager(self._asset)

    def reserve_address(self) -> MaybeValue:
        """Get the reserve address for the asset."""
        return AssetParam.reserve(self._asset)

    def freeze_address(self) -> MaybeValue:
        """Get the freeze address for the asset."""
        return AssetParam.freeze(self._asset)

    def clawback_address(self) -> MaybeValue:
        """Get the clawback address for the asset."""
        return AssetParam.clawback(self._asset)

    def creator_address(self) -> MaybeValue:
        """Get the creator address for the asset."""
        return AssetParam.creator(self._asset)


AssetParamObject.__module__ = "pyteal"



import re
from enum import Enum
from typing import Any

from pyteal.errors import TealTypeError, TealInputError


class TealType(Enum):
    """Teal type enum."""

    """Unsigned 64 bit integer type."""
    uint64 = 0

    """Byte string type."""
    bytes = 1

    """Any type that is not none."""
    anytype = 2

    """Represents no value."""
    none = 3


TealType.__module__ = "pyteal"


def require_type(input: Any, expected: TealType):
    try:
        actual = input.type_of()
    except AttributeError:
        raise TypeError(f"Expected a {expected} object, but got a {type(input)}")

    if actual != expected and (
        expected == TealType.none
        or actual == TealType.none
        or (actual != TealType.anytype and expected != TealType.anytype)
    ):
        raise TealTypeError(actual, expected)


def types_match(type1: TealType, type2: TealType) -> bool:
    if (type1 == TealType.none or type2 == TealType.none) and type1 != type2:
        return False

    if type1 == TealType.anytype or type2 == TealType.anytype:
        return True

    return type1 == type2


def valid_address(address: str):
    """check if address is a valid address with checksum"""
    if type(address) is not str:
        raise TealInputError("An address needs to be a string")

    if len(address) != 58:
        raise TealInputError(
            "Address length is not correct. Should "
            + "be a base 32 string encoded 32 bytes public key + 4 bytes checksum"
        )

    valid_base32(address)


def valid_base32(s: str):
    """check if s is a valid base32 encoding string"""
    pattern = re.compile(
        r"^(?:[A-Z2-7]{8})*(?:([A-Z2-7]{2}([=]{6})?)|([A-Z2-7]{4}([=]{4})?)|([A-Z2-7]{5}([=]{3})?)|([A-Z2-7]{7}([=]{1})?))?"
    )

    if pattern.fullmatch(s) is None:
        raise TealInputError("{} is not a valid RFC 4648 base 32 string".format(s))


def valid_base64(s: str):
    """check if s is a valid base64 encoding string"""
    pattern = re.compile(
        r"^(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$"
    )

    if pattern.fullmatch(s) is None:
        raise TealInputError("{} is not a valid RFC 4648 base 64 string".format(s))


def valid_base16(s: str):
    """check if s is a valid hex encoding string"""
    if len(s) % 2:
        raise TealInputError(
            f"{s} is of odd length, not a valid RFC 4648 base16 string"
        )

    pattern = re.compile(r"[0-9A-Fa-f]*")

    if pattern.fullmatch(s) is None:
        raise TealInputError("{} is not a valid RFC 4648 base 16 string".format(s))


def valid_tmpl(s: str):
    """check if s is valid template name"""
    pattern = re.compile(r"TMPL_[A-Z0-9_]+")

    if pattern.fullmatch(s) is None:
        raise TealInputError("{} is not a valid template variable".format(s))



from typing import TYPE_CHECKING
from enum import Enum

from pyteal.types import TealType, require_type
from pyteal.errors import verifyFieldVersion
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.ast.expr import Expr
from pyteal.ast.leafexpr import LeafExpr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class BlockField(Enum):
    # fmt: off
    #                 id  |    name     |      type     | min version
    block_seed =      (0, "BlkSeed",      TealType.bytes,  7)  # noqa: E222
    block_timestamp = (1, "BlkTimestamp", TealType.uint64, 7)  # noqa: E222

    # fmt: on

    def __init__(self, id: int, name: str, type: TealType, min_version: int) -> None:
        self.id = id
        self.arg_name = name
        self.ret_type = type
        self.min_version = min_version

    def type_of(self) -> TealType:
        return self.ret_type


BlockField.__module__ = "pyteal"


class Block(LeafExpr):
    """An expression that accesses a block property."""

    def __init__(self, field: BlockField, block: Expr) -> None:
        super().__init__()
        self.field = field

        require_type(block, TealType.uint64)
        self.block = block

    def __teal__(self, options: "CompileOptions"):
        verifyFieldVersion(self.field.arg_name, self.field.min_version, options.version)

        op = TealOp(self, Op.block, self.field.arg_name)
        return TealBlock.FromOp(options, op, self.block)

    def __str__(self):
        return "(Block {})".format(self.field.arg_name)

    def type_of(self):
        return self.field.type_of()

    @classmethod
    def seed(cls, block: Expr) -> Expr:
        """Get the seed of a block.

        Args:
            block: A block index that corresponds to the block to check,
                must be evaluated to uint64. Fails if the block index is not less than
                :code:`Txn.first_valid()` or more than 1001 rounds before :code:`Txn.last_valid()`.
        """
        return cls(BlockField.block_seed, block)

    @classmethod
    def timestamp(cls, block: Expr) -> Expr:
        """Get the timestamp of a block.

        Args:
            block: A block index that corresponds to the block to check,
                must be evaluated to uint64. Fails if the block index is not less than
                :code:`Txn.first_valid()` or more than 1001 rounds before :code:`Txn.last_valid()`.
        """
        return cls(BlockField.block_timestamp, block)


Block.__module__ = "pyteal"



from typing import TYPE_CHECKING

from pyteal.types import TealType, valid_tmpl
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.ast.leafexpr import LeafExpr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class Tmpl(LeafExpr):
    """Template expression for creating placeholder values."""

    def __init__(self, op: Op, type: TealType, name: str) -> None:
        super().__init__()
        valid_tmpl(name)
        self.op = op
        self.type = type
        self.name = name

    def __str__(self):
        return "(Tmpl {} {})".format(self.op, self.name)

    def __teal__(self, options: "CompileOptions"):
        op = TealOp(self, self.op, self.name)
        return TealBlock.FromOp(options, op)

    def type_of(self):
        return self.type

    @classmethod
    def Int(cls, placeholder: str):
        """Create a new Int template.

        Args:
            placeholder: The name to use for this template variable. Must start with `TMPL_` and
                only consist of uppercase alphanumeric characters and underscores.
        """
        return cls(Op.int, TealType.uint64, placeholder)

    @classmethod
    def Bytes(cls, placeholder: str):
        """Create a new Bytes template.

        Args:
            placeholder: The name to use for this template variable. Must start with `TMPL_` and
                only consist of uppercase alphanumeric characters and underscores.
        """
        return cls(Op.byte, TealType.bytes, placeholder)

    @classmethod
    def Addr(cls, placeholder: str):
        """Create a new Addr template.

        Args:
            placeholder: The name to use for this template variable. Must start with `TMPL_` and
                only consist of uppercase alphanumeric characters and underscores.
        """
        return cls(Op.addr, TealType.bytes, placeholder)


Tmpl.__module__ = "pyteal"



from typing import TYPE_CHECKING

from pyteal.types import TealType
from pyteal.errors import TealCompileError
from pyteal.ast.expr import Expr
from pyteal.ir import TealSimpleBlock


if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class Break(Expr):
    """A break expression"""

    def __init__(self) -> None:
        """Create a new break expression.

        This operation is only permitted in a loop.

        """
        super().__init__()

    def __str__(self) -> str:
        return "break"

    def __teal__(self, options: "CompileOptions"):
        if not options.isInLoop():
            raise TealCompileError("break is only allowed in a loop", self)

        start = TealSimpleBlock([])
        options.addLoopBreakBlock(start)

        return start, start

    def type_of(self):
        return TealType.none

    def has_return(self):
        return False


Break.__module__ = "pyteal"



import pytest

import pyteal as pt
from pyteal.types import require_type


def test_require_type():
    require_type(pt.Bytes("str"), pt.TealType.bytes)
    assert True


def test_require_type_invalid():
    with pytest.raises(TypeError):
        pt.App.globalGet(["This is certainly invalid"])



# This example is provided for informational purposes only and has not been audited for security.

from pyteal import *

"""Atomic Swap"""

alice = Addr("6ZHGHH5Z5CTPCF5WCESXMGRSVK7QJETR63M3NY5FJCUYDHO57VTCMJOBGY")
bob = Addr("7Z5PWO2C6LFNQFGHWKSK5H47IQP5OJW2M3HA2QPXTY3WTNP5NU2MHBW27M")
secret = Bytes("base32", "2323232323232323")
timeout = 3000


def htlc(
    tmpl_seller=alice,
    tmpl_buyer=bob,
    tmpl_fee=1000,
    tmpl_secret=secret,
    tmpl_hash_fn=Sha256,
    tmpl_timeout=timeout,
):

    fee_cond = Txn.fee() < Int(tmpl_fee)
    safety_cond = And(
        Txn.type_enum() == TxnType.Payment,
        Txn.close_remainder_to() == Global.zero_address(),
        Txn.rekey_to() == Global.zero_address(),
    )

    recv_cond = And(Txn.receiver() == tmpl_seller, tmpl_hash_fn(Arg(0)) == tmpl_secret)

    esc_cond = And(Txn.receiver() == tmpl_buyer, Txn.first_valid() > Int(tmpl_timeout))

    return And(fee_cond, safety_cond, Or(recv_cond, esc_cond))


if __name__ == "__main__":
    print(compileTeal(htlc(), mode=Mode.Signature, version=2))



import pytest

import pyteal as pt

avm3Options = pt.CompileOptions(version=3)
avm4Options = pt.CompileOptions(version=4)
avm6Options = pt.CompileOptions(version=6)


def test_gload_teal_3():
    with pytest.raises(pt.TealInputError):
        pt.ImportScratchValue(0, 1).__teal__(avm3Options)

    with pytest.raises(pt.TealInputError):
        pt.ImportScratchValue(pt.Int(0), 1).__teal__(avm3Options)

    with pytest.raises(pt.TealInputError):
        pt.ImportScratchValue(pt.Int(0), pt.Int(1)).__teal__(avm3Options)


def test_gload_teal_4():
    with pytest.raises(pt.TealInputError):
        pt.ImportScratchValue(pt.Int(0), pt.Int(2)).__teal__(avm4Options)


def test_gload():
    expr = pt.ImportScratchValue(0, 1)
    assert expr.type_of() == pt.TealType.anytype

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.gload, 0, 1)])

    actual, _ = expr.__teal__(avm4Options)

    assert actual == expected


def test_gloads():
    arg = pt.Int(1)
    expr = pt.ImportScratchValue(arg, 0)
    assert expr.type_of() == pt.TealType.anytype

    expected = pt.TealSimpleBlock(
        [pt.TealOp(arg, pt.Op.int, 1), pt.TealOp(expr, pt.Op.gloads, 0)]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_gloadss():
    txID = pt.Int(1)
    slotID = pt.Int(0)
    expr = pt.ImportScratchValue(txID, slotID)
    assert expr.type_of() == pt.TealType.anytype

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(txID, pt.Op.int, 1),
            pt.TealOp(slotID, pt.Op.int, 0),
            pt.TealOp(expr, pt.Op.gloadss),
        ]
    )

    actual, _ = expr.__teal__(avm6Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_gload_invalid():
    with pytest.raises(pt.TealInputError):
        pt.ImportScratchValue(-1, 0)

    with pytest.raises(pt.TealInputError):
        pt.ImportScratchValue(pt.MAX_GROUP_SIZE, 0)

    with pytest.raises(pt.TealInputError):
        pt.ImportScratchValue(0, -1)

    with pytest.raises(pt.TealInputError):
        pt.ImportScratchValue(0, pt.NUM_SLOTS)

    with pytest.raises(pt.TealInputError):
        pt.ImportScratchValue(0, pt.Int(0))

    with pytest.raises(pt.TealTypeError):
        pt.ImportScratchValue(pt.Bytes("AQID"), 0)  # byte encoding of [1, 2, 3]



import pytest

import pyteal as pt

avm2Options = pt.CompileOptions(version=2)
avm3Options = pt.CompileOptions(version=3)
avm4Options = pt.CompileOptions(version=4)
avm5Options = pt.CompileOptions(version=5)
avm6Options = pt.CompileOptions(version=6)
avm7Options = pt.CompileOptions(version=7)


def test_ed25519verify():
    args = [pt.Bytes("data"), pt.Bytes("sig"), pt.Bytes("key")]
    expr = pt.Ed25519Verify(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"data"'),
            pt.TealOp(args[1], pt.Op.byte, '"sig"'),
            pt.TealOp(args[2], pt.Op.byte, '"key"'),
            pt.TealOp(expr, pt.Op.ed25519verify),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_ed25519verify_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Ed25519Verify(pt.Int(0), pt.Bytes("sig"), pt.Bytes("key"))

    with pytest.raises(pt.TealTypeError):
        pt.Ed25519Verify(pt.Bytes("data"), pt.Int(0), pt.Bytes("key"))

    with pytest.raises(pt.TealTypeError):
        pt.Ed25519Verify(pt.Bytes("data"), pt.Bytes("sig"), pt.Int(0))


def test_ed25519verify_bare():
    args = [pt.Bytes("data"), pt.Bytes("sig"), pt.Bytes("key")]
    expr = pt.Ed25519Verify_Bare(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"data"'),
            pt.TealOp(args[1], pt.Op.byte, '"sig"'),
            pt.TealOp(args[2], pt.Op.byte, '"key"'),
            pt.TealOp(expr, pt.Op.ed25519verify_bare),
        ]
    )

    actual, _ = expr.__teal__(avm7Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm6Options)


def test_ed25519verify_bare_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Ed25519Verify_Bare(pt.Int(0), pt.Bytes("sig"), pt.Bytes("key"))

    with pytest.raises(pt.TealTypeError):
        pt.Ed25519Verify_Bare(pt.Bytes("data"), pt.Int(0), pt.Bytes("key"))

    with pytest.raises(pt.TealTypeError):
        pt.Ed25519Verify_Bare(pt.Bytes("data"), pt.Bytes("sig"), pt.Int(0))


def test_set_bit_int():
    args = [pt.Int(0), pt.Int(2), pt.Int(1)]
    expr = pt.SetBit(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 0),
            pt.TealOp(args[1], pt.Op.int, 2),
            pt.TealOp(args[2], pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.setbit),
        ]
    )

    actual, _ = expr.__teal__(avm3Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm2Options)


def test_set_bit_bytes():
    args = [pt.Bytes("base16", "0x0000"), pt.Int(0), pt.Int(1)]
    expr = pt.SetBit(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, "0x0000"),
            pt.TealOp(args[1], pt.Op.int, 0),
            pt.TealOp(args[2], pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.setbit),
        ]
    )

    actual, _ = expr.__teal__(avm3Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm2Options)


def test_set_bit_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.SetBit(pt.Int(3), pt.Bytes("index"), pt.Int(1))

    with pytest.raises(pt.TealTypeError):
        pt.SetBit(pt.Int(3), pt.Int(0), pt.Bytes("one"))

    with pytest.raises(pt.TealTypeError):
        pt.SetBit(pt.Bytes("base16", "0xFF"), pt.Bytes("index"), pt.Int(1))

    with pytest.raises(pt.TealTypeError):
        pt.SetBit(pt.Bytes("base16", "0xFF"), pt.Int(0), pt.Bytes("one"))


def test_set_byte():
    args = [pt.Bytes("base16", "0xFF"), pt.Int(0), pt.Int(3)]
    expr = pt.SetByte(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, "0xFF"),
            pt.TealOp(args[1], pt.Op.int, 0),
            pt.TealOp(args[2], pt.Op.int, 3),
            pt.TealOp(expr, pt.Op.setbyte),
        ]
    )

    actual, _ = expr.__teal__(avm3Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm2Options)


def test_set_byte_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.SetByte(pt.Int(3), pt.Int(0), pt.Int(1))

    with pytest.raises(pt.TealTypeError):
        pt.SetByte(pt.Bytes("base16", "0xFF"), pt.Bytes("index"), pt.Int(1))

    with pytest.raises(pt.TealTypeError):
        pt.SetByte(pt.Bytes("base16", "0xFF"), pt.Int(0), pt.Bytes("one"))


def test_divw():
    args = [pt.Int(0), pt.Int(90), pt.Int(30)]
    expr = pt.Divw(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, args[0].value),
            pt.TealOp(args[1], pt.Op.int, args[1].value),
            pt.TealOp(args[2], pt.Op.int, args[2].value),
            pt.TealOp(expr, pt.Op.divw),
        ]
    )

    actual, _ = expr.__teal__(avm6Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_divw_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Divw(pt.Bytes("10"), pt.Int(0), pt.Int(1))

    with pytest.raises(pt.TealTypeError):
        pt.Divw(pt.Int(10), pt.Bytes("0"), pt.Int(1))

    with pytest.raises(pt.TealTypeError):
        pt.Divw(pt.Int(10), pt.Int(0), pt.Bytes("1"))



from typing import TYPE_CHECKING

from pyteal.errors import TealInputError
from pyteal.ast.expr import Expr
from pyteal.ast.seq import Seq
from pyteal.ast.bytes import Bytes
from pyteal.ast.unaryexpr import Pop

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class Nonce(Expr):
    """A meta expression only used to change the hash of a TEAL program."""

    def __init__(self, base: str, nonce: str, child: Expr) -> None:
        """Create a new Nonce.

        The Nonce expression behaves exactly like the child expression passed into it, except it
        uses the provided nonce string to alter its structure in a way that does not affect
        execution.

        Args:
            base: The base of the nonce. Must be one of utf8, base16, base32, or base64.
            nonce: An arbitrary nonce string that conforms to base.
            child: The expression to wrap.
        """
        super().__init__()

        if base not in ("utf8", "base16", "base32", "base64"):
            raise TealInputError("Invalid base: {}".format(base))

        self.child = child
        if base == "utf8":
            self.nonce_bytes = Bytes(nonce)
        else:
            self.nonce_bytes = Bytes(base, nonce)

        self.seq = Seq([Pop(self.nonce_bytes), self.child])

    def __teal__(self, options: "CompileOptions"):
        return self.seq.__teal__(options)

    def __str__(self):
        return "(nonce: {}) {}".format(self.nonce_bytes, self.child)

    def type_of(self):
        return self.child.type_of()

    def has_return(self):
        return self.child.has_return()


Nonce.__module__ = "pyteal"



import pytest
from tests.mock_version import mock_version  # noqa: F401

import pyteal as pt


@pytest.mark.usefixtures("mock_version")
@pytest.mark.parametrize(
    "version, compiler_version, should_error",
    [
        # valid
        ("0.12.0", "0.12.0", False),
        (
            "1.0.0+AVM7.1",
            "=1.0.0",
            False,
        ),
        # invalid
        ("0.13.0", "0.13.1", True),
        ("1.2.3a2", "<0.8.0 || >=0.12.0", True),
    ],
)
def test_pragma_expr(compiler_version, should_error):
    program = pt.Pragma(pt.Approve(), compiler_version=compiler_version)

    if should_error:
        with pytest.raises(pt.TealPragmaError):
            pt.compileTeal(program, mode=pt.Mode.Application, version=6)
    else:
        pt.compileTeal(program, mode=pt.Mode.Application, version=6)


def test_pragma_expr_does_not_change():
    without_pragma = pt.Seq(pt.Pop(pt.Add(pt.Int(1), pt.Int(2))), pt.Return(pt.Int(1)))
    pragma = pt.Pragma(without_pragma, compiler_version=">=0.0.0")

    compiled_with_pragma = pt.compileTeal(pragma, mode=pt.Mode.Application, version=6)
    compiled_without_pragma = pt.compileTeal(
        without_pragma, mode=pt.Mode.Application, version=6
    )

    assert compiled_with_pragma == compiled_without_pragma


def test_pragma_expr_has_return():
    exprWithReturn = pt.Pragma(pt.Return(pt.Int(1)), compiler_version=">=0.0.0")
    assert exprWithReturn.has_return()

    exprWithoutReturn = pt.Pragma(pt.Int(1), compiler_version=">=0.0.0")
    assert not exprWithoutReturn.has_return()


@pytest.mark.parametrize(
    "compiler_version",
    ["not a version", ">=0.1.1,<0.3.0", "1.2.3aq"],  # incorrect spec  # invalid PEP 440
)
def test_pragma_expr_invalid_compiler_version(compiler_version):
    with pytest.raises(ValueError):
        pt.Pragma(pt.Approve(), compiler_version=compiler_version)



import pyteal as pt
import secrets
from pyteal.ast.router import ASTBuilder
import pytest
import typing
import algosdk.abi as sdk_abi


options = pt.CompileOptions(version=5)


@pt.ABIReturnSubroutine
def add(a: pt.abi.Uint64, b: pt.abi.Uint64, *, output: pt.abi.Uint64) -> pt.Expr:
    """add takes 2 integers a,b and adds them, returning the sum"""
    return output.set(a.get() + b.get())


@pt.ABIReturnSubroutine
def sub(a: pt.abi.Uint64, b: pt.abi.Uint64, *, output: pt.abi.Uint64) -> pt.Expr:
    """replace me"""
    return output.set(a.get() - b.get())


@pt.ABIReturnSubroutine
def mul(a: pt.abi.Uint64, b: pt.abi.Uint64, *, output: pt.abi.Uint64) -> pt.Expr:
    return output.set(a.get() * b.get())


@pt.ABIReturnSubroutine
def div(a: pt.abi.Uint64, b: pt.abi.Uint64, *, output: pt.abi.Uint64) -> pt.Expr:
    return output.set(a.get() / b.get())


@pt.ABIReturnSubroutine
def mod(a: pt.abi.Uint64, b: pt.abi.Uint64, *, output: pt.abi.Uint64) -> pt.Expr:
    return output.set(a.get() % b.get())


@pt.ABIReturnSubroutine
def qrem(
    a: pt.abi.Uint64,
    b: pt.abi.Uint64,
    *,
    output: pt.abi.Tuple2[pt.abi.Uint64, pt.abi.Uint64],
) -> pt.Expr:
    return pt.Seq(
        (q := pt.abi.Uint64()).set(a.get() / b.get()),
        (rem := pt.abi.Uint64()).set(a.get() % b.get()),
        output.set(q, rem),
    )


@pt.ABIReturnSubroutine
def reverse(a: pt.abi.String, *, output: pt.abi.String) -> pt.Expr:
    idx = pt.ScratchVar()
    buff = pt.ScratchVar()

    init = idx.store(pt.Int(0))
    cond = idx.load() < a.length()
    _iter = idx.store(idx.load() + pt.Int(1))
    return pt.Seq(
        buff.store(pt.Bytes("")),
        pt.For(init, cond, _iter).Do(
            a[idx.load()].use(lambda v: buff.store(pt.Concat(v.encode(), buff.load())))
        ),
        output.set(buff.load()),
    )


@pt.ABIReturnSubroutine
def concat_strings(
    b: pt.abi.DynamicArray[pt.abi.String], *, output: pt.abi.String
) -> pt.Expr:
    idx = pt.ScratchVar()
    buff = pt.ScratchVar()

    init = idx.store(pt.Int(0))
    cond = idx.load() < b.length()
    _iter = idx.store(idx.load() + pt.Int(1))
    return pt.Seq(
        buff.store(pt.Bytes("")),
        pt.For(init, cond, _iter).Do(
            b[idx.load()].use(lambda s: buff.store(pt.Concat(buff.load(), s.get())))
        ),
        output.set(buff.load()),
    )


@pt.ABIReturnSubroutine
def many_args(
    _a: pt.abi.Uint64,
    _b: pt.abi.Uint64,
    _c: pt.abi.Uint64,
    _d: pt.abi.Uint64,
    _e: pt.abi.Uint64,
    _f: pt.abi.Uint64,
    _g: pt.abi.Uint64,
    _h: pt.abi.Uint64,
    _i: pt.abi.Uint64,
    _j: pt.abi.Uint64,
    _k: pt.abi.Uint64,
    _l: pt.abi.Uint64,
    _m: pt.abi.Uint64,
    _n: pt.abi.Uint64,
    _o: pt.abi.Uint64,
    _p: pt.abi.Uint64,
    _q: pt.abi.Uint64,
    _r: pt.abi.Uint64,
    _s: pt.abi.Uint64,
    _t: pt.abi.Uint64,
    *,
    output: pt.abi.Uint64,
) -> pt.Expr:
    return output.set(_t.get())


@pt.ABIReturnSubroutine
def many_args_with_transaction(
    _txn: pt.abi.Transaction,
    _b: pt.abi.Uint64,
    _c: pt.abi.Uint64,
    _d: pt.abi.Uint64,
    _e: pt.abi.Uint64,
    _f: pt.abi.Uint64,
    _g: pt.abi.Uint64,
    _h: pt.abi.Uint64,
    _i: pt.abi.Uint64,
    _j: pt.abi.Uint64,
    _k: pt.abi.Uint64,
    _l: pt.abi.Uint64,
    _m: pt.abi.Uint64,
    _n: pt.abi.Uint64,
    _o: pt.abi.Uint64,
    _p: pt.abi.Uint64,
    _q: pt.abi.Uint64,
    _r: pt.abi.Uint64,
    _s: pt.abi.Uint64,
    _t: pt.abi.Uint64,
    *,
    output: pt.abi.Uint64,
) -> pt.Expr:
    return output.set(_t.get())


@pt.ABIReturnSubroutine
def add_and_store(
    a1: pt.abi.Uint64,
    a2: pt.abi.Uint64,
    a3: pt.abi.Uint64,
    a4: pt.abi.Uint64,
    a5: pt.abi.Uint64,
    a6: pt.abi.Uint64,
    a7: pt.abi.Uint64,
    a8: pt.abi.Uint64,
    a9: pt.abi.Uint64,
    a10: pt.abi.Uint64,
    a11: pt.abi.Uint64,
    a12: pt.abi.Uint64,
    a13: pt.abi.Uint64,
    a14: pt.abi.Uint64,
    a15: pt.abi.Uint64,
    a16: pt.abi.Uint64,
    t1: pt.abi.PaymentTransaction,
    *,
    output: pt.abi.Uint64,
) -> pt.Expr:
    return pt.Seq(
        output.set(a1.get() + a2.get()),
        # store the result in the sender's local state too
        pt.App.localPut(pt.Txn.sender(), pt.Bytes("result"), output.get()),
    )


@pt.Subroutine(pt.TealType.none)
def safe_clear_state_delete():
    return (
        pt.If(pt.Txn.sender() == pt.Global.creator_address())
        .Then(pt.Approve())
        .Else(pt.Reject())
    )


@pt.ABIReturnSubroutine
def dummy_doing_nothing():
    return pt.Seq(pt.Log(pt.Bytes("a message")))


@pt.Subroutine(pt.TealType.uint64)
def returning_u64():
    return pt.Int(1)


@pt.Subroutine(pt.TealType.none)
def mult_over_u64_and_log(a: pt.Expr, b: pt.Expr):
    return pt.Log(pt.Itob(a * b))


@pt.ABIReturnSubroutine
def eine_constant(*, output: pt.abi.Uint64):
    return output.set(1)


@pt.ABIReturnSubroutine
def take_abi_and_log(tb_logged: pt.abi.String):
    return pt.Log(tb_logged.get())


@pt.ABIReturnSubroutine
def not_registrable(lhs: pt.abi.Uint64, rhs: pt.Expr, *, output: pt.abi.Uint64):
    return output.set(lhs.get() * rhs)


@pt.ABIReturnSubroutine
def txn_amount(t: pt.abi.PaymentTransaction, *, output: pt.abi.Uint64):
    return output.set(t.get().amount())


@pt.ABIReturnSubroutine
def multiple_txn(
    appl: pt.abi.ApplicationCallTransaction,
    axfer: pt.abi.AssetTransferTransaction,
    pay: pt.abi.PaymentTransaction,
    any_txn: pt.abi.Transaction,
    *,
    output: pt.abi.Uint64,
):
    return output.set(
        appl.get().fee() + axfer.get().fee() + pay.get().fee() + any_txn.get().fee()
    )


GOOD_SUBROUTINE_CASES: list[pt.ABIReturnSubroutine | pt.SubroutineFnWrapper] = [
    add,
    sub,
    mul,
    div,
    mod,
    qrem,
    reverse,
    concat_strings,
    many_args,
    many_args_with_transaction,
    add_and_store,
    safe_clear_state_delete,
    dummy_doing_nothing,
    eine_constant,
    take_abi_and_log,
    txn_amount,
    multiple_txn,
]

ON_COMPLETE_CASES: list[pt.EnumInt] = [
    pt.OnComplete.NoOp,
    pt.OnComplete.OptIn,
    pt.OnComplete.ClearState,
    pt.OnComplete.CloseOut,
    pt.OnComplete.UpdateApplication,
    pt.OnComplete.DeleteApplication,
]


def power_set(no_dup_list: list, length_override: int = None):
    """
    This function serves as a generator for all possible elements in power_set
    over `non_dup_list`, which is a list of non-duplicated elements (matches property of a set).

    The cardinality of a powerset is 2^|non_dup_list|, so we can iterate from 0 to 2^|non_dup_list| - 1
    to index each element in such power_set.
    By binary representation of each index, we can see it as an allowance over each element in `no_dup_list`,
    and generate a unique subset of `non_dup_list`, which yields as an element of power_set of `no_dup_list`.

    Args:
        no_dup_list: a list of elements with no duplication
        length_override: a number indicating the largest size of super_set element,
            must be in range [1, len(no_dup_list)].
    """
    if length_override is None:
        length_override = len(no_dup_list)
    assert 1 <= length_override <= len(no_dup_list)
    masks = [1 << i for i in range(length_override)]
    for i in range(1 << len(no_dup_list)):
        yield [elem for mask, elem in zip(masks, no_dup_list) if i & mask]


class FullOrderCombinationGen:
    """
    This class serves as a generator for all possible vectors of maximal length `largest_perm_length` (non-negative),
    each of whose entries are one of the elements in `non_dup_list`, namely, a list of non-duplicated elements.
    """

    def __init__(self, non_dup_list: list, largest_perm_length: int) -> None:
        if largest_perm_length < 0:
            raise pt.TealInputError(
                "largest input permutation length must be non-negative"
            )
        elif len(set(non_dup_list)) != len(non_dup_list):
            raise pt.TealInputError(
                f"input non_dup_list {non_dup_list} has duplications"
            )
        elif not len(non_dup_list):
            raise pt.TealInputError("input non_dup_list must be non empty")

        self.__basis_symbol = non_dup_list
        self.__basis_size = len(self.__basis_symbol)
        self.__pre_gen_table: list[list[int]] = [
            [] for _ in range(self.__basis_size**largest_perm_length)
        ]

        # we can index all possible cases of vectors with an index in range
        # [0, |non_dup_list| ^ perm_length - 1]
        # by converting an index into |non_dup_list|-based number,
        # we can get the vector mapped by the index.

        # we iterate through [0, |non_dup_list|^largest_perm_length - 1] to precompute permutation table.
        lhs_scope = 0
        discrete_log = 0
        for expn in range(largest_perm_length + 1):
            for index in range(lhs_scope, self.__basis_size**expn):
                basis_repr = [0 for _ in range(discrete_log)]
                if discrete_log:
                    temp = index
                    for i in range(discrete_log):
                        basis_repr[i] = temp % self.__basis_size
                        temp //= self.__basis_size
                self.__pre_gen_table[index] = basis_repr

            lhs_scope = self.__basis_size**expn
            discrete_log = expn + 1

    def sample_gen(self, perm_length: int, sample_num: int = 10):
        if perm_length < 0:
            raise pt.TealInputError("input permutation length must be non-negative")
        elif perm_length == 0:
            yield []
            return

        # since we are sampling for a permutation with length `perm_length`,
        # this corresponds to sampling a value from [|non_dup_list|^(perm_length - 1), |non_dup_list|^perm_length - 1].
        # if sample number is greater than interval size, by pigeonhole principle there is re-testing
        # reduce back down to interval size
        sample_num = min(sample_num, self.__basis_size ** (perm_length - 1))

        for _ in range(sample_num):
            take = secrets.choice(
                range(
                    self.__basis_size ** (perm_length - 1),
                    self.__basis_size**perm_length,
                )
            )
            yield [self.__basis_symbol[j] for j in self.__pre_gen_table[take]]


def assemble_helper(what: pt.Expr) -> pt.TealBlock:
    assembled, _ = what.__teal__(options)
    assembled.addIncoming()
    assembled = pt.TealBlock.NormalizeBlocks(assembled)
    return assembled


def camel_to_snake(name: str) -> str:
    return "".join(["_" + c.lower() if c.isupper() else c for c in name]).lstrip("_")


def test_call_config():
    for cc in pt.CallConfig:
        approval_cond_on_cc: pt.Expr | int = cc.approval_condition_under_config()
        match approval_cond_on_cc:
            case pt.Expr():
                expected_cc = (
                    (pt.Txn.application_id() == pt.Int(0))
                    if cc == pt.CallConfig.CREATE
                    else (pt.Txn.application_id() != pt.Int(0))
                )
                with pt.TealComponent.Context.ignoreExprEquality():
                    assert assemble_helper(approval_cond_on_cc) == assemble_helper(
                        expected_cc
                    )
            case int():
                assert approval_cond_on_cc == int(cc) & 1
            case _:
                raise pt.TealInternalError(
                    f"unexpected approval_cond_on_cc {approval_cond_on_cc}"
                )

        if cc in (pt.CallConfig.CREATE, pt.CallConfig.ALL):
            with pytest.raises(
                pt.TealInputError,
                match=r"Only CallConfig.CALL or CallConfig.NEVER are valid for a clear state CallConfig, since clear state can never be invoked during creation$",
            ):
                cc.clear_state_condition_under_config()
            continue

        clear_state_cond_on_cc: int = cc.clear_state_condition_under_config()
        match clear_state_cond_on_cc:
            case 0:
                assert cc == pt.CallConfig.NEVER
            case 1:
                assert cc == pt.CallConfig.CALL
            case _:
                raise pt.TealInternalError(
                    f"unexpected clear_state_cond_on_cc {clear_state_cond_on_cc}"
                )


def test_method_config_call_config_never():
    never_mc = pt.MethodConfig(no_op=pt.CallConfig.NEVER)
    assert never_mc.is_never()
    assert never_mc.approval_cond() == 0
    assert never_mc.clear_state_cond() == 0


def _gen_method_configs(sample_count: int = 10):
    on_complete_pow_set = power_set(ON_COMPLETE_CASES)
    focg = FullOrderCombinationGen(list(pt.CallConfig), len(ON_COMPLETE_CASES))

    for on_complete_set in on_complete_pow_set:
        oc_names = [camel_to_snake(oc.name) for oc in on_complete_set]
        for call_configs in focg.sample_gen(len(on_complete_set), sample_count):
            yield pt.MethodConfig(**dict(zip(oc_names, call_configs)))


@pytest.mark.parametrize("mc", _gen_method_configs())
def test_method_config(mc: pt.MethodConfig):
    approval_check_names_n_ocs = [
        (camel_to_snake(oc.name), oc)
        for oc in ON_COMPLETE_CASES
        if str(oc) != str(pt.OnComplete.ClearState)
    ]

    match mc.clear_state:
        case pt.CallConfig.NEVER:
            assert mc.clear_state_cond() == 0
        case pt.CallConfig.CALL:
            assert mc.clear_state_cond() == 1
        case pt.CallConfig.CREATE | pt.CallConfig.ALL:
            with pytest.raises(
                pt.TealInputError,
                match=r"Only CallConfig.CALL or CallConfig.NEVER are valid for a clear state CallConfig, since clear state can never be invoked during creation$",
            ):
                mc.clear_state_cond()
    if mc.is_never() or all(
        getattr(mc, i) == pt.CallConfig.NEVER for i, _ in approval_check_names_n_ocs
    ):
        assert mc.approval_cond() == 0
        return
    elif all(
        getattr(mc, i) == pt.CallConfig.ALL for i, _ in approval_check_names_n_ocs
    ):
        assert mc.approval_cond() == 1
        return
    list_of_cc = [
        (
            typing.cast(
                pt.CallConfig, getattr(mc, i)
            ).approval_condition_under_config(),
            oc,
        )
        for i, oc in approval_check_names_n_ocs
    ]
    list_of_expressions: list[pt.Expr] = []
    for expr_or_int, oc in list_of_cc:
        match expr_or_int:
            case pt.Expr():
                list_of_expressions.append(
                    pt.And(pt.Txn.on_completion() == oc, expr_or_int)
                )
            case 0:
                continue
            case 1:
                list_of_expressions.append(pt.Txn.on_completion() == oc)
    with pt.TealComponent.Context.ignoreExprEquality():
        ac = mc.approval_cond()
        assert isinstance(ac, pt.Expr)
        assert assemble_helper(ac) == assemble_helper(pt.Or(*list_of_expressions))


def test_on_complete_action():
    with pytest.raises(pt.TealInputError) as contradict_err:
        pt.OnCompleteAction(action=pt.Seq(), call_config=pt.CallConfig.NEVER)
    assert "contradicts" in str(contradict_err)
    assert pt.OnCompleteAction.never().is_empty()
    assert pt.OnCompleteAction.call_only(pt.Seq()).call_config == pt.CallConfig.CALL
    assert pt.OnCompleteAction.create_only(pt.Seq()).call_config == pt.CallConfig.CREATE
    assert pt.OnCompleteAction.always(pt.Seq()).call_config == pt.CallConfig.ALL


def test_wrap_handler_bare_call():
    BARE_CALL_CASES = [
        dummy_doing_nothing,
        safe_clear_state_delete,
        pt.Approve(),
        pt.Log(pt.Bytes("message")),
    ]
    for bare_call in BARE_CALL_CASES:
        wrapped: pt.Expr = ASTBuilder.wrap_handler(False, bare_call)
        expected: pt.Expr
        match bare_call:
            case pt.Expr():
                if bare_call.has_return():
                    expected = bare_call
                else:
                    expected = pt.Seq(bare_call, pt.Approve())
            case pt.SubroutineFnWrapper() | pt.ABIReturnSubroutine():
                expected = pt.Seq(bare_call(), pt.Approve())
            case _:
                raise pt.TealInputError("how you got here?")
        wrapped_assemble = assemble_helper(wrapped)
        wrapped_helper = assemble_helper(expected)
        with pt.TealComponent.Context.ignoreExprEquality():
            assert wrapped_assemble == wrapped_helper

    ERROR_CASES = [
        (
            pt.Int(1),
            f"bare appcall handler should be TealType.none not {pt.TealType.uint64}.",
        ),
        (
            returning_u64,
            f"subroutine call should be returning TealType.none not {pt.TealType.uint64}.",
        ),
        (
            mult_over_u64_and_log,
            "subroutine call should take 0 arg for bare-app call. this subroutine takes 2.",
        ),
        (
            eine_constant,
            f"abi-returning subroutine call should be returning void not {pt.abi.Uint64TypeSpec()}.",
        ),
        (
            take_abi_and_log,
            "abi-returning subroutine call should take 0 arg for bare-app call. this abi-returning subroutine takes 1.",
        ),
        (
            1,
            "bare appcall can only accept: none type Expr, or Subroutine/ABIReturnSubroutine with none return and no arg",
        ),
    ]
    for error_case, error_msg in ERROR_CASES:
        with pytest.raises(pt.TealInputError) as bug:
            ASTBuilder.wrap_handler(False, error_case)
        assert error_msg in str(bug)


def test_wrap_handler_method_call():
    with pytest.raises(pt.TealInputError) as bug:
        ASTBuilder.wrap_handler(True, not_registrable)
    assert "method call ABIReturnSubroutine is not routable" in str(bug)

    with pytest.raises(pt.TealInputError) as bug:
        ASTBuilder.wrap_handler(True, safe_clear_state_delete)
    assert "method call should be only registering ABIReturnSubroutine" in str(bug)

    ONLY_ABI_SUBROUTINE_CASES = list(
        filter(lambda x: isinstance(x, pt.ABIReturnSubroutine), GOOD_SUBROUTINE_CASES)
    )

    for abi_subroutine in ONLY_ABI_SUBROUTINE_CASES:
        wrapped: pt.Expr = ASTBuilder.wrap_handler(True, abi_subroutine)
        actual: pt.TealBlock = assemble_helper(wrapped)

        args: list[pt.abi.BaseType] = [
            spec.new_instance()
            for spec in typing.cast(
                list[pt.abi.TypeSpec], abi_subroutine.subroutine.expected_arg_types
            )
        ]

        app_args = [
            arg for arg in args if arg.type_spec() not in pt.abi.TransactionTypeSpecs
        ]

        app_arg_cnt = len(app_args)

        txn_args: list[pt.abi.Transaction] = [
            arg for arg in args if arg.type_spec() in pt.abi.TransactionTypeSpecs
        ]

        loading: list[pt.Expr] = []

        if app_arg_cnt > pt.METHOD_ARG_NUM_CUTOFF:
            sdk_last_arg = pt.abi.TupleTypeSpec(
                *[arg.type_spec() for arg in app_args[pt.METHOD_ARG_NUM_CUTOFF - 1 :]]
            ).new_instance()

            loading = [
                arg.decode(pt.Txn.application_args[index + 1])
                for index, arg in enumerate(app_args[: pt.METHOD_ARG_NUM_CUTOFF - 1])
            ]

            loading.append(
                sdk_last_arg.decode(pt.Txn.application_args[pt.METHOD_ARG_NUM_CUTOFF])
            )
        else:
            loading = [
                arg.decode(pt.Txn.application_args[index + 1])
                for index, arg in enumerate(app_args)
            ]

        if len(txn_args) > 0:
            for idx, txn_arg in enumerate(txn_args):
                loading.append(
                    txn_arg._set_index(
                        pt.Txn.group_index() - pt.Int(len(txn_args) - idx)
                    )
                )
                if str(txn_arg.type_spec()) != "txn":
                    loading.append(
                        pt.Assert(
                            txn_arg.get().type_enum()
                            == txn_arg.type_spec().txn_type_enum()
                        )
                    )

        if app_arg_cnt > pt.METHOD_ARG_NUM_CUTOFF:
            loading.extend(
                [
                    sdk_last_arg[idx].store_into(val)
                    for idx, val in enumerate(app_args[pt.METHOD_ARG_NUM_CUTOFF - 1 :])
                ]
            )

        evaluate: pt.Expr
        if abi_subroutine.type_of() != "void":
            output_temp = abi_subroutine.output_kwarg_info.abi_type.new_instance()
            evaluate = pt.Seq(
                abi_subroutine(*args).store_into(output_temp),
                pt.abi.MethodReturn(output_temp),
            )
        else:
            evaluate = abi_subroutine(*args)

        expected = assemble_helper(pt.Seq(*loading, evaluate, pt.Approve()))
        with pt.TealComponent.Context.ignoreScratchSlotEquality(), pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected

        assert pt.TealBlock.MatchScratchSlotReferences(
            pt.TealBlock.GetReferencedScratchSlots(actual),
            pt.TealBlock.GetReferencedScratchSlots(expected),
        )


def test_wrap_handler_method_txn_types():
    wrapped: pt.Expr = ASTBuilder.wrap_handler(True, multiple_txn)
    actual: pt.TealBlock = assemble_helper(wrapped)

    args: list[pt.abi.Transaction] = [
        pt.abi.ApplicationCallTransaction(),
        pt.abi.AssetTransferTransaction(),
        pt.abi.PaymentTransaction(),
        pt.abi.Transaction(),
    ]
    output_temp = pt.abi.Uint64()
    expected_ast = pt.Seq(
        args[0]._set_index(pt.Txn.group_index() - pt.Int(4)),
        pt.Assert(args[0].get().type_enum() == pt.TxnType.ApplicationCall),
        args[1]._set_index(pt.Txn.group_index() - pt.Int(3)),
        pt.Assert(args[1].get().type_enum() == pt.TxnType.AssetTransfer),
        args[2]._set_index(pt.Txn.group_index() - pt.Int(2)),
        pt.Assert(args[2].get().type_enum() == pt.TxnType.Payment),
        args[3]._set_index(pt.Txn.group_index() - pt.Int(1)),
        multiple_txn(*args).store_into(output_temp),
        pt.abi.MethodReturn(output_temp),
        pt.Approve(),
    )

    expected = assemble_helper(expected_ast)
    with pt.TealComponent.Context.ignoreScratchSlotEquality(), pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    assert pt.TealBlock.MatchScratchSlotReferences(
        pt.TealBlock.GetReferencedScratchSlots(actual),
        pt.TealBlock.GetReferencedScratchSlots(expected),
    )


def test_wrap_handler_method_call_many_args():
    wrapped: pt.Expr = ASTBuilder.wrap_handler(True, many_args)
    actual: pt.TealBlock = assemble_helper(wrapped)

    args = [pt.abi.Uint64() for _ in range(20)]
    last_arg = pt.abi.TupleTypeSpec(
        *[pt.abi.Uint64TypeSpec() for _ in range(6)]
    ).new_instance()

    output_temp = pt.abi.Uint64()
    expected_ast = pt.Seq(
        args[0].decode(pt.Txn.application_args[1]),
        args[1].decode(pt.Txn.application_args[2]),
        args[2].decode(pt.Txn.application_args[3]),
        args[3].decode(pt.Txn.application_args[4]),
        args[4].decode(pt.Txn.application_args[5]),
        args[5].decode(pt.Txn.application_args[6]),
        args[6].decode(pt.Txn.application_args[7]),
        args[7].decode(pt.Txn.application_args[8]),
        args[8].decode(pt.Txn.application_args[9]),
        args[9].decode(pt.Txn.application_args[10]),
        args[10].decode(pt.Txn.application_args[11]),
        args[11].decode(pt.Txn.application_args[12]),
        args[12].decode(pt.Txn.application_args[13]),
        args[13].decode(pt.Txn.application_args[14]),
        last_arg.decode(pt.Txn.application_args[15]),
        last_arg[0].store_into(args[14]),
        last_arg[1].store_into(args[15]),
        last_arg[2].store_into(args[16]),
        last_arg[3].store_into(args[17]),
        last_arg[4].store_into(args[18]),
        last_arg[5].store_into(args[19]),
        many_args(*args).store_into(output_temp),
        pt.abi.MethodReturn(output_temp),
        pt.Approve(),
    )
    expected = assemble_helper(expected_ast)
    with pt.TealComponent.Context.ignoreScratchSlotEquality(), pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    assert pt.TealBlock.MatchScratchSlotReferences(
        pt.TealBlock.GetReferencedScratchSlots(actual),
        pt.TealBlock.GetReferencedScratchSlots(expected),
    )


def test_contract_json_obj():
    abi_subroutines = list(
        filter(lambda x: isinstance(x, pt.ABIReturnSubroutine), GOOD_SUBROUTINE_CASES)
    )
    contract_name = "contract_name"
    on_complete_actions = pt.BareCallActions(
        clear_state=pt.OnCompleteAction.call_only(safe_clear_state_delete)
    )
    router = pt.Router(contract_name, on_complete_actions)
    method_list: list[sdk_abi.Method] = []
    for subroutine in abi_subroutines:

        doc = subroutine.subroutine.implementation.__doc__
        desc = None
        if doc is not None and doc.strip() == "replace me":
            desc = "dope description"

        router.add_method_handler(subroutine, description=desc)

        ms = subroutine.method_spec()

        # Manually replace it since the override is applied in the method handler
        # not attached to the ABIReturnSubroutine itself
        ms.desc = desc if desc is not None else ms.desc

        sig_method = sdk_abi.Method.from_signature(subroutine.method_signature())

        assert ms.name == sig_method.name

        for idx, arg in enumerate(ms.args):
            assert arg.type == sig_method.args[idx].type

        method_list.append(ms)

    sdk_contract = sdk_abi.Contract(contract_name, method_list)
    contract = router.contract_construct()
    assert contract == sdk_contract


def test_build_program_all_empty():
    router = pt.Router("test")

    approval, clear_state, contract = router.build_program()

    expected_empty_program = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.return_),
        ]
    )

    with pt.TealComponent.Context.ignoreExprEquality():
        assert assemble_helper(approval) == expected_empty_program
        assert assemble_helper(clear_state) == expected_empty_program

    expected_contract = sdk_abi.Contract("test", [])
    assert contract == expected_contract


def test_build_program_approval_empty():
    router = pt.Router(
        "test",
        pt.BareCallActions(clear_state=pt.OnCompleteAction.call_only(pt.Approve())),
    )

    approval, clear_state, contract = router.build_program()

    expected_empty_program = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.return_),
        ]
    )

    with pt.TealComponent.Context.ignoreExprEquality():
        assert assemble_helper(approval) == expected_empty_program
        assert assemble_helper(clear_state) != expected_empty_program

    expected_contract = sdk_abi.Contract("test", [])
    assert contract == expected_contract


def test_build_program_clear_state_empty():
    router = pt.Router(
        "test", pt.BareCallActions(no_op=pt.OnCompleteAction.always(pt.Approve()))
    )

    approval, clear_state, contract = router.build_program()

    expected_empty_program = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.return_),
        ]
    )

    with pt.TealComponent.Context.ignoreExprEquality():
        assert assemble_helper(approval) != expected_empty_program
        assert assemble_helper(clear_state) == expected_empty_program

    expected_contract = sdk_abi.Contract("test", [])
    assert contract == expected_contract


def test_build_program_clear_state_invalid_config():
    for config in (pt.CallConfig.CREATE, pt.CallConfig.ALL):
        bareCalls = pt.BareCallActions(
            clear_state=pt.OnCompleteAction(action=pt.Approve(), call_config=config)
        )
        with pytest.raises(
            pt.TealInputError,
            match=r"Only CallConfig.CALL or CallConfig.NEVER are valid for a clear state CallConfig, since clear state can never be invoked during creation$",
        ):
            pt.Router("test", bareCalls)

        router = pt.Router("test")

        @pt.ABIReturnSubroutine
        def clear_state_method():
            return pt.Approve()

        with pytest.raises(
            pt.TealInputError,
            match=r"Only CallConfig.CALL or CallConfig.NEVER are valid for a clear state CallConfig, since clear state can never be invoked during creation$",
        ):
            router.add_method_handler(
                clear_state_method,
                method_config=pt.MethodConfig(clear_state=config),
            )


def test_build_program_clear_state_valid_config():
    action = pt.If(pt.Txn.fee() == pt.Int(4)).Then(pt.Approve()).Else(pt.Reject())
    config = pt.CallConfig.CALL

    router_with_bare_call = pt.Router(
        "test",
        pt.BareCallActions(
            clear_state=pt.OnCompleteAction(action=action, call_config=config)
        ),
    )
    _, actual_clear_state_with_bare_call, _ = router_with_bare_call.build_program()

    expected_clear_state_with_bare_call = assemble_helper(
        pt.Cond([pt.Txn.application_args.length() == pt.Int(0), action])
    )

    with pt.TealComponent.Context.ignoreExprEquality():
        assert (
            assemble_helper(actual_clear_state_with_bare_call)
            == expected_clear_state_with_bare_call
        )

    router_with_method = pt.Router("test")

    @pt.ABIReturnSubroutine
    def clear_state_method():
        return action

    router_with_method.add_method_handler(
        clear_state_method, method_config=pt.MethodConfig(clear_state=config)
    )

    _, actual_clear_state_with_method, _ = router_with_method.build_program()

    expected_clear_state_with_method = assemble_helper(
        pt.Cond(
            [
                pt.Txn.application_args[0]
                == pt.MethodSignature("clear_state_method()void"),
                pt.Seq(clear_state_method(), pt.Approve()),
            ]
        )
    )

    with pt.TealComponent.Context.ignoreExprEquality():
        assert (
            assemble_helper(actual_clear_state_with_method)
            == expected_clear_state_with_method
        )


def test_override_names():
    r1 = pt.Router("test")

    @r1.method(name="handle")
    def handle_asa(deposit: pt.abi.AssetTransferTransaction):
        """handles the deposit where the input is an asset transfer"""
        return pt.Assert(deposit.get().asset_amount() > pt.Int(0))

    @r1.method(name="handle")
    def handle_algo(deposit: pt.abi.PaymentTransaction):
        """handles the deposit where the input is a payment"""
        return pt.Assert(deposit.get().amount() > pt.Int(0))

    ap1, cs1, c1 = r1.compile_program(version=pt.compiler.MAX_PROGRAM_VERSION)
    assert len(c1.methods) == 2
    for meth in c1.methods:
        dmeth = meth.dictify()
        assert dmeth["name"] == "handle"

    # Confirm an equivalent router definition _without_ `name` overrides produces the same output.
    r2 = pt.Router("test")

    @r2.method()
    def handle(deposit: pt.abi.AssetTransferTransaction):
        """handles the deposit where the input is an asset transfer"""
        return pt.Assert(deposit.get().asset_amount() > pt.Int(0))

    @r2.method()
    def handle(deposit: pt.abi.PaymentTransaction):  # noqa: F811
        """handles the deposit where the input is a payment"""
        return pt.Assert(deposit.get().amount() > pt.Int(0))

    ap2, cs2, c2 = r2.compile_program(version=pt.compiler.MAX_PROGRAM_VERSION)

    assert (ap1, cs1, c1) == (ap2, cs2, c2)



import pytest

import pyteal as pt
from tests.compile_asserts import assert_new_v_old


def user_guide_snippet_dynamic_scratch_var() -> pt.Expr:
    """
    The user guide docs use the test to illustrate `DynamicScratchVar` usage.  If the test breaks, then the user guide docs must be updated along with the test.
    """
    from pyteal import Assert, Int, DynamicScratchVar, ScratchVar, Seq, TealType

    s = ScratchVar(TealType.uint64)
    d = DynamicScratchVar(TealType.uint64)

    return Seq(
        d.set_index(s),
        s.store(Int(7)),
        d.store(d.load() + Int(3)),
        Assert(s.load() == Int(10)),
        Int(1),
    )


@pytest.mark.parametrize("snippet", [user_guide_snippet_dynamic_scratch_var])
def test_user_guide_snippets(snippet):
    assert_new_v_old(snippet, 6, "user_guide")


def user_guide_snippet_recursiveIsEven():
    from pyteal import If, Int, Subroutine, TealType

    @Subroutine(TealType.uint64)
    def recursiveIsEven(i):
        return (
            If(i == Int(0))
            .Then(Int(1))
            .ElseIf(i == Int(1))
            .Then(Int(0))
            .Else(recursiveIsEven(i - Int(2)))
        )

    return recursiveIsEven(Int(15))


def user_guide_snippet_ILLEGAL_recursion():
    from pyteal import If, Int, ScratchVar, Seq, Subroutine, TealType

    @Subroutine(TealType.none)
    def ILLEGAL_recursion(i: ScratchVar):
        return (
            If(i.load() == Int(0))
            .Then(i.store(Int(1)))
            .ElseIf(i.load() == Int(1))
            .Then(i.store(Int(0)))
            .Else(Seq(i.store(i.load() - Int(2)), ILLEGAL_recursion(i)))
        )

    i = ScratchVar(TealType.uint64)
    return Seq(i.store(Int(15)), ILLEGAL_recursion(i), Int(1))


def user_guide_snippet_ABIReturnSubroutine():
    from pyteal import (
        ABIReturnSubroutine,
        Expr,
        For,
        Int,
        ScratchVar,
        Seq,
        Txn,
        TealType,
    )
    from pyteal import abi

    # --- BEGIN doc-comment --- #
    @ABIReturnSubroutine
    def abi_sum(to_sum: abi.DynamicArray[abi.Uint64], *, output: abi.Uint64) -> Expr:
        i = ScratchVar(TealType.uint64)
        value_at_index = abi.Uint64()
        return Seq(
            output.set(0),
            For(
                i.store(Int(0)), i.load() < to_sum.length(), i.store(i.load() + Int(1))
            ).Do(
                Seq(
                    to_sum[i.load()].store_into(value_at_index),
                    output.set(output.get() + value_at_index.get()),
                )
            ),
        )

    program = Seq(
        (to_sum_arr := abi.make(abi.DynamicArray[abi.Uint64])).decode(
            Txn.application_args[1]
        ),
        (res := abi.Uint64()).set(abi_sum(to_sum_arr)),
        pt.abi.MethodReturn(res),
        Int(1),
    )
    # --- END doc-comment --- #
    return program


USER_GUIDE_SNIPPETS_COPACETIC = [
    user_guide_snippet_dynamic_scratch_var,
    user_guide_snippet_recursiveIsEven,
    user_guide_snippet_ABIReturnSubroutine,
]


@pytest.mark.parametrize("snippet", USER_GUIDE_SNIPPETS_COPACETIC)
def test_user_guide_snippets_good(snippet):
    assert_new_v_old(snippet, 6, "user_guide")


USER_GUIDE_SNIPPETS_ERRORING = {
    user_guide_snippet_ILLEGAL_recursion: (
        pt.TealInputError,
        "ScratchVar arguments not allowed in recursive subroutines, but a recursive call-path was detected: ILLEGAL_recursion()-->ILLEGAL_recursion()",
    )
}


@pytest.mark.parametrize("snippet_etype_e", USER_GUIDE_SNIPPETS_ERRORING.items())
def test_user_guide_snippets_bad(snippet_etype_e):
    snippet, etype_e = snippet_etype_e
    etype, e = etype_e

    print(
        f"Test case function=[{snippet.__name__}]. Expecting error of type {etype} with message <{e}>"
    )
    with pytest.raises(etype) as tie:
        pt.compileTeal(snippet(), mode=pt.Mode.Application, version=6)

    assert e in str(tie)



from typing import Union, cast, TYPE_CHECKING

from pyteal.types import TealType, require_type
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.errors import TealInputError, verifyProgramVersion
from pyteal.ast.expr import Expr
from pyteal.ast.leafexpr import LeafExpr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class Arg(LeafExpr):
    """An expression to get an argument when running in signature verification mode."""

    def __init__(self, index: Union[int, Expr]) -> None:
        """Get an argument for this program.

        Should only be used in signature verification mode. For application mode arguments, see
        :any:`TxnObject.application_args`.

        Args:
            index: The index of the argument to get. The index must be between 0 and 255 inclusive.
                Starting in AVM v5, the index may be a PyTeal expression that evaluates to uint64.
        """
        super().__init__()

        if type(index) is int:
            if index < 0 or index > 255:
                raise TealInputError("invalid arg index {}".format(index))
        else:
            require_type(cast(Expr, index), TealType.uint64)

        self.index = index

    def __teal__(self, options: "CompileOptions"):
        if type(self.index) is int:
            op = TealOp(self, Op.arg, self.index)
            return TealBlock.FromOp(options, op)

        verifyProgramVersion(
            Op.args.min_version,
            options.version,
            "Program version too low to use dynamic indexes with Arg",
        )

        op = TealOp(self, Op.args)
        return TealBlock.FromOp(options, op, cast(Expr, self.index))

    def __str__(self):
        return "(arg {})".format(self.index)

    def type_of(self):
        return TealType.bytes


Arg.__module__ = "pyteal"



from typing import Final, TypeVar, cast
from abc import abstractmethod
from pyteal.ast.abi.type import BaseType, TypeSpec
from pyteal.ast.abi.uint import NUM_BITS_IN_BYTE, uint_decode

from pyteal.ast.expr import Expr
from pyteal.ast.txn import Txn
from pyteal.ast.acct import AccountParamObject
from pyteal.ast.asset import AssetHoldingObject, AssetParamObject
from pyteal.ast.app import AppParamObject
from pyteal.errors import TealInputError
from pyteal.types import TealType


T = TypeVar("T", bound="ReferenceType")


class ReferenceTypeSpec(TypeSpec):
    @abstractmethod
    def new_instance(self) -> "ReferenceType":
        pass

    @abstractmethod
    def annotation_type(self) -> "type[ReferenceType]":
        pass

    def bit_size(self) -> int:
        """Get the bit size of the index this reference type holds"""
        return NUM_BITS_IN_BYTE

    def is_dynamic(self) -> bool:
        return False

    def byte_length_static(self) -> int:
        return 1

    def storage_type(self) -> TealType:
        return TealType.uint64


ReferenceTypeSpec.__module__ = "pyteal.abi"


class ReferenceType(BaseType):
    @abstractmethod
    def __init__(self, spec: ReferenceTypeSpec) -> None:
        super().__init__(spec)

    def type_spec(self) -> ReferenceTypeSpec:
        return cast(ReferenceTypeSpec, super().type_spec())

    def referenced_index(self) -> Expr:
        """Get the reference index for this value.

        The three reference types (account, application, asset) contain indexes into a foreign array
        of the transaction. This method returns that index.

        If this reference type is an application or asset, note that this DOES NOT return the
        application or asset ID. See :code:`application_id()` or :code:`asset_id()` for that.
        """
        return self.stored_value.load()

    def decode(
        self,
        encoded: Expr,
        *,
        start_index: Expr = None,
        end_index: Expr = None,
        length: Expr = None,
    ) -> Expr:
        return uint_decode(
            self.type_spec().bit_size(),
            self.stored_value,
            encoded,
            start_index,
            end_index,
            length,
        )

    def encode(self) -> Expr:
        raise TealInputError("A ReferenceType cannot be encoded")


ReferenceType.__module__ = "pyteal.abi"


class AccountTypeSpec(ReferenceTypeSpec):
    def new_instance(self) -> "Account":
        return Account()

    def annotation_type(self) -> "type[Account]":
        return Account

    def __str__(self) -> str:
        return "account"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, AccountTypeSpec)


AccountTypeSpec.__module__ = "pyteal.abi"


class Account(ReferenceType):
    def __init__(self) -> None:
        super().__init__(AccountTypeSpec())

    def address(self) -> Expr:
        """Get the address of the account."""
        return Txn.accounts[self.stored_value.load()]

    def params(self) -> AccountParamObject:
        """Get information about the account."""
        return AccountParamObject(self.referenced_index())

    def asset_holding(self, asset: "Expr | Asset") -> AssetHoldingObject:
        """Get information about an asset held by this account.

        Args:
            asset: An identifier for the asset. It must be one of the following: an abi.Asset
                reference object, an expression holding an index into Txn.ForeignAssets that
                corresponds to the asset (in which case it must evaluate to uint64), or since v4, an
                expression holding an asset ID that appears in Txn.ForeignAssets (in which case it
                must evaluate to uint64).
        """
        if isinstance(asset, Asset):
            asset_ref = asset.referenced_index()
        else:
            asset_ref = asset
        return AssetHoldingObject(asset_ref, self.referenced_index())


Account.__module__ = "pyteal.abi"


class AssetTypeSpec(ReferenceTypeSpec):
    def new_instance(self) -> "Asset":
        return Asset()

    def annotation_type(self) -> "type[Asset]":
        return Asset

    def __str__(self) -> str:
        return "asset"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, AssetTypeSpec)


AssetTypeSpec.__module__ = "pyteal.abi"


class Asset(ReferenceType):
    def __init__(self) -> None:
        super().__init__(AssetTypeSpec())

    def asset_id(self) -> Expr:
        """Get the ID of the asset."""
        return Txn.assets[self.referenced_index()]

    def holding(self, account: Expr | Account) -> AssetHoldingObject:
        """Get information about this asset held by an account.

        Args:
            account: An identifier for the account. It must be one of the following: an abi.Account
                reference object, an expression holding an index into Txn.Accounts that corresponds
                to the account (in which case it must evaluate to uint64), or since v4, an
                expression holding an account address that appears in Txn.Accounts or is Txn.Sender
                (in which case it must evaluate to bytes).
        """
        if isinstance(account, Account):
            account_ref = account.referenced_index()
        else:
            account_ref = account
        return AssetHoldingObject(self.referenced_index(), account_ref)

    def params(self) -> AssetParamObject:
        """Get information about the asset's parameters."""
        return AssetParamObject(self.referenced_index())


Asset.__module__ = "pyteal.abi"


class ApplicationTypeSpec(ReferenceTypeSpec):
    def new_instance(self) -> "Application":
        return Application()

    def annotation_type(self) -> "type[Application]":
        return Application

    def __str__(self) -> str:
        return "application"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ApplicationTypeSpec)


ApplicationTypeSpec.__module__ = "pyteal.abi"


class Application(ReferenceType):
    def __init__(self) -> None:
        super().__init__(ApplicationTypeSpec())

    def application_id(self) -> Expr:
        """Get the ID of the application."""
        return Txn.applications[self.stored_value.load()]

    def params(self) -> AppParamObject:
        """Get information about the application's parameters."""
        return AppParamObject(self.referenced_index())


Application.__module__ = "pyteal.abi"


ReferenceTypeSpecs: Final[list[TypeSpec]] = [
    AccountTypeSpec(),
    AssetTypeSpec(),
    ApplicationTypeSpec(),
]



import pytest

import pyteal as pt

from pyteal.compiler.scratchslots import (
    collectScratchSlots,
    assignScratchSlotsToSubroutines,
)


def test_collectScratchSlots():
    def sub1Impl():
        return None

    def sub2Impl(a1):
        return None

    def sub3Impl(a1, a2, a3):
        return None

    subroutine1 = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)
    subroutine2 = pt.SubroutineDefinition(sub2Impl, pt.TealType.bytes)
    subroutine3 = pt.SubroutineDefinition(sub3Impl, pt.TealType.none)

    globalSlot1 = pt.ScratchSlot()

    subroutine1Slot1 = pt.ScratchSlot()
    subroutine1Slot2 = pt.ScratchSlot()
    subroutine1Ops = [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.store, subroutine1Slot1),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.store, subroutine1Slot2),
        pt.TealOp(None, pt.Op.load, globalSlot1),
        pt.TealOp(None, pt.Op.retsub),
    ]

    subroutine2Slot1 = pt.ScratchSlot()
    subroutine2Ops = [
        pt.TealOp(None, pt.Op.byte, '"value"'),
        pt.TealOp(None, pt.Op.store, subroutine2Slot1),
        pt.TealOp(None, pt.Op.load, subroutine2Slot1),
        pt.TealOp(None, pt.Op.retsub),
    ]

    subroutine3Ops = [
        pt.TealOp(None, pt.Op.retsub),
    ]

    mainSlot1 = pt.ScratchSlot()
    mainSlot2 = pt.ScratchSlot()
    mainOps = [
        pt.TealOp(None, pt.Op.int, 7),
        pt.TealOp(None, pt.Op.store, globalSlot1),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.store, mainSlot1),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.store, mainSlot2),
        pt.TealOp(None, pt.Op.load, mainSlot1),
        pt.TealOp(None, pt.Op.return_),
    ]

    subroutineBlocks = {
        None: pt.TealSimpleBlock(mainOps),
        subroutine1: pt.TealSimpleBlock(subroutine1Ops),
        subroutine2: pt.TealSimpleBlock(subroutine2Ops),
        subroutine3: pt.TealSimpleBlock(subroutine3Ops),
    }

    expected_global = {globalSlot1}

    expected_local = {
        None: {mainSlot1, mainSlot2},
        subroutine1: {subroutine1Slot1, subroutine1Slot2},
        subroutine2: {subroutine2Slot1},
        subroutine3: set(),
    }

    actual_global, actual_local = collectScratchSlots(subroutineBlocks)

    assert actual_global == expected_global
    assert actual_local == expected_local


def test_assignScratchSlotsToSubroutines_no_requested_ids():
    def sub1Impl():
        return None

    def sub2Impl(a1):
        return None

    def sub3Impl(a1, a2, a3):
        return None

    subroutine1 = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)
    subroutine2 = pt.SubroutineDefinition(sub2Impl, pt.TealType.bytes)
    subroutine3 = pt.SubroutineDefinition(sub3Impl, pt.TealType.none)

    globalSlot1 = pt.ScratchSlot()

    subroutine1Slot1 = pt.ScratchSlot()
    subroutine1Slot2 = pt.ScratchSlot()
    subroutine1Ops = [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.store, subroutine1Slot1),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.store, subroutine1Slot2),
        pt.TealOp(None, pt.Op.load, globalSlot1),
        pt.TealOp(None, pt.Op.retsub),
    ]

    subroutine2Slot1 = pt.ScratchSlot()
    subroutine2Ops = [
        pt.TealOp(None, pt.Op.byte, '"value"'),
        pt.TealOp(None, pt.Op.store, subroutine2Slot1),
        pt.TealOp(None, pt.Op.load, subroutine2Slot1),
        pt.TealOp(None, pt.Op.retsub),
    ]

    subroutine3Ops = [
        pt.TealOp(None, pt.Op.retsub),
    ]

    mainSlot1 = pt.ScratchSlot()
    mainSlot2 = pt.ScratchSlot()
    mainOps = [
        pt.TealOp(None, pt.Op.int, 7),
        pt.TealOp(None, pt.Op.store, globalSlot1),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.store, mainSlot1),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.store, mainSlot2),
        pt.TealOp(None, pt.Op.load, mainSlot1),
        pt.TealOp(None, pt.Op.return_),
    ]

    subroutineBlocks = {
        None: pt.TealSimpleBlock(mainOps),
        subroutine1: pt.TealSimpleBlock(subroutine1Ops),
        subroutine2: pt.TealSimpleBlock(subroutine2Ops),
        subroutine3: pt.TealSimpleBlock(subroutine3Ops),
    }

    expectedAssignments = {
        globalSlot1: 0,
        subroutine1Slot1: 1,
        subroutine1Slot2: 2,
        subroutine2Slot1: 3,
        mainSlot1: 4,
        mainSlot2: 5,
    }

    expected = {
        None: {expectedAssignments[mainSlot1], expectedAssignments[mainSlot2]},
        subroutine1: {
            expectedAssignments[subroutine1Slot1],
            expectedAssignments[subroutine1Slot2],
        },
        subroutine2: {expectedAssignments[subroutine2Slot1]},
        subroutine3: set(),
    }

    actual = assignScratchSlotsToSubroutines(subroutineBlocks)

    assert actual == expected

    assert subroutine1Ops == [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.store, expectedAssignments[subroutine1Slot1]),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.store, expectedAssignments[subroutine1Slot2]),
        pt.TealOp(None, pt.Op.load, expectedAssignments[globalSlot1]),
        pt.TealOp(None, pt.Op.retsub),
    ]

    assert subroutine2Ops == [
        pt.TealOp(None, pt.Op.byte, '"value"'),
        pt.TealOp(None, pt.Op.store, expectedAssignments[subroutine2Slot1]),
        pt.TealOp(None, pt.Op.load, expectedAssignments[subroutine2Slot1]),
        pt.TealOp(None, pt.Op.retsub),
    ]

    assert subroutine3Ops == [
        pt.TealOp(None, pt.Op.retsub),
    ]

    assert mainOps == [
        pt.TealOp(None, pt.Op.int, 7),
        pt.TealOp(None, pt.Op.store, expectedAssignments[globalSlot1]),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.store, expectedAssignments[mainSlot1]),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.store, expectedAssignments[mainSlot2]),
        pt.TealOp(None, pt.Op.load, expectedAssignments[mainSlot1]),
        pt.TealOp(None, pt.Op.return_),
    ]


def test_assignScratchSlotsToSubroutines_with_requested_ids():
    def sub1Impl():
        return None

    def sub2Impl(a1):
        return None

    def sub3Impl(a1, a2, a3):
        return None

    subroutine1 = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)
    subroutine2 = pt.SubroutineDefinition(sub2Impl, pt.TealType.bytes)
    subroutine3 = pt.SubroutineDefinition(sub3Impl, pt.TealType.none)

    globalSlot1 = pt.ScratchSlot(requestedSlotId=8)

    subroutine1Slot1 = pt.ScratchSlot()
    subroutine1Slot2 = pt.ScratchSlot(requestedSlotId=5)
    subroutine1Ops = [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.store, subroutine1Slot1),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.store, subroutine1Slot2),
        pt.TealOp(None, pt.Op.load, globalSlot1),
        pt.TealOp(None, pt.Op.retsub),
    ]

    subroutine2Slot1 = pt.ScratchSlot()
    subroutine2Ops = [
        pt.TealOp(None, pt.Op.byte, '"value"'),
        pt.TealOp(None, pt.Op.store, subroutine2Slot1),
        pt.TealOp(None, pt.Op.load, subroutine2Slot1),
        pt.TealOp(None, pt.Op.retsub),
    ]

    subroutine3Ops = [
        pt.TealOp(None, pt.Op.retsub),
    ]

    mainSlot1 = pt.ScratchSlot()
    mainSlot2 = pt.ScratchSlot(requestedSlotId=100)
    mainOps = [
        pt.TealOp(None, pt.Op.int, 7),
        pt.TealOp(None, pt.Op.store, globalSlot1),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.store, mainSlot1),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.store, mainSlot2),
        pt.TealOp(None, pt.Op.load, mainSlot1),
        pt.TealOp(None, pt.Op.return_),
    ]

    subroutineBlocks = {
        None: pt.TealSimpleBlock(mainOps),
        subroutine1: pt.TealSimpleBlock(subroutine1Ops),
        subroutine2: pt.TealSimpleBlock(subroutine2Ops),
        subroutine3: pt.TealSimpleBlock(subroutine3Ops),
    }

    expectedAssignments = {
        globalSlot1: 8,
        subroutine1Slot1: 0,
        subroutine1Slot2: 5,
        subroutine2Slot1: 1,
        mainSlot1: 2,
        mainSlot2: 100,
    }

    expected = {
        None: {expectedAssignments[mainSlot1], expectedAssignments[mainSlot2]},
        subroutine1: {
            expectedAssignments[subroutine1Slot1],
            expectedAssignments[subroutine1Slot2],
        },
        subroutine2: {expectedAssignments[subroutine2Slot1]},
        subroutine3: set(),
    }

    actual = assignScratchSlotsToSubroutines(subroutineBlocks)

    assert actual == expected

    assert subroutine1Ops == [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.store, expectedAssignments[subroutine1Slot1]),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.store, expectedAssignments[subroutine1Slot2]),
        pt.TealOp(None, pt.Op.load, expectedAssignments[globalSlot1]),
        pt.TealOp(None, pt.Op.retsub),
    ]

    assert subroutine2Ops == [
        pt.TealOp(None, pt.Op.byte, '"value"'),
        pt.TealOp(None, pt.Op.store, expectedAssignments[subroutine2Slot1]),
        pt.TealOp(None, pt.Op.load, expectedAssignments[subroutine2Slot1]),
        pt.TealOp(None, pt.Op.retsub),
    ]

    assert subroutine3Ops == [
        pt.TealOp(None, pt.Op.retsub),
    ]

    assert mainOps == [
        pt.TealOp(None, pt.Op.int, 7),
        pt.TealOp(None, pt.Op.store, expectedAssignments[globalSlot1]),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.store, expectedAssignments[mainSlot1]),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.store, expectedAssignments[mainSlot2]),
        pt.TealOp(None, pt.Op.load, expectedAssignments[mainSlot1]),
        pt.TealOp(None, pt.Op.return_),
    ]


def test_assignScratchSlotsToSubroutines_invalid_requested_id():
    def sub1Impl():
        return None

    def sub2Impl(a1):
        return None

    def sub3Impl(a1, a2, a3):
        return None

    subroutine1 = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)
    subroutine2 = pt.SubroutineDefinition(sub2Impl, pt.TealType.bytes)
    subroutine3 = pt.SubroutineDefinition(sub3Impl, pt.TealType.none)

    globalSlot1 = pt.ScratchSlot(requestedSlotId=8)

    subroutine1Slot1 = pt.ScratchSlot()
    subroutine1Slot2 = pt.ScratchSlot(requestedSlotId=5)
    subroutine1Ops = [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.store, subroutine1Slot1),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.store, subroutine1Slot2),
        pt.TealOp(None, pt.Op.load, globalSlot1),
        pt.TealOp(None, pt.Op.retsub),
    ]

    subroutine2Slot1 = pt.ScratchSlot(requestedSlotId=100)
    subroutine2Ops = [
        pt.TealOp(None, pt.Op.byte, '"value"'),
        pt.TealOp(None, pt.Op.store, subroutine2Slot1),
        pt.TealOp(None, pt.Op.load, subroutine2Slot1),
        pt.TealOp(None, pt.Op.retsub),
    ]

    subroutine3Ops = [
        pt.TealOp(None, pt.Op.retsub),
    ]

    mainSlot1 = pt.ScratchSlot()
    mainSlot2 = pt.ScratchSlot(requestedSlotId=100)
    mainOps = [
        pt.TealOp(None, pt.Op.int, 7),
        pt.TealOp(None, pt.Op.store, globalSlot1),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.store, mainSlot1),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.store, mainSlot2),
        pt.TealOp(None, pt.Op.load, mainSlot1),
        pt.TealOp(None, pt.Op.return_),
    ]

    subroutineBlocks = {
        None: pt.TealSimpleBlock(mainOps),
        subroutine1: pt.TealSimpleBlock(subroutine1Ops),
        subroutine2: pt.TealSimpleBlock(subroutine2Ops),
        subroutine3: pt.TealSimpleBlock(subroutine3Ops),
    }

    # mainSlot2 and subroutine2Slot1 request the same ID, 100
    with pytest.raises(pt.TealInternalError):
        assignScratchSlotsToSubroutines(subroutineBlocks)


def test_assignScratchSlotsToSubroutines_slot_used_before_assignment():
    def sub1Impl():
        return None

    def sub2Impl(a1):
        return None

    def sub3Impl(a1, a2, a3):
        return None

    subroutine1 = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)
    subroutine2 = pt.SubroutineDefinition(sub2Impl, pt.TealType.bytes)
    subroutine3 = pt.SubroutineDefinition(sub3Impl, pt.TealType.none)

    globalSlot1 = pt.ScratchSlot()

    subroutine1Slot1 = pt.ScratchSlot()
    subroutine1Slot2 = pt.ScratchSlot()
    subroutine1Ops = [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.store, subroutine1Slot1),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.store, subroutine1Slot2),
        pt.TealOp(None, pt.Op.load, globalSlot1),
        pt.TealOp(None, pt.Op.retsub),
    ]

    subroutine2Slot1 = pt.ScratchSlot()
    subroutine2Ops = [
        pt.TealOp(None, pt.Op.byte, '"value"'),
        pt.TealOp(None, pt.Op.store, subroutine2Slot1),
        pt.TealOp(None, pt.Op.load, subroutine2Slot1),
        pt.TealOp(None, pt.Op.retsub),
    ]

    subroutine3Ops = [
        pt.TealOp(None, pt.Op.retsub),
    ]

    mainSlot1 = pt.ScratchSlot()
    mainSlot2 = pt.ScratchSlot()
    mainOps = [
        pt.TealOp(None, pt.Op.int, 7),
        pt.TealOp(None, pt.Op.store, globalSlot1),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.store, mainSlot2),
        pt.TealOp(None, pt.Op.load, mainSlot1),
        pt.TealOp(None, pt.Op.return_),
    ]

    subroutineBlocks = {
        None: pt.TealSimpleBlock(mainOps),
        subroutine1: pt.TealSimpleBlock(subroutine1Ops),
        subroutine2: pt.TealSimpleBlock(subroutine2Ops),
        subroutine3: pt.TealSimpleBlock(subroutine3Ops),
    }

    with pytest.raises(pt.TealInternalError):
        assignScratchSlotsToSubroutines(subroutineBlocks)



from typing import TYPE_CHECKING

from pyteal.errors import TealInputError
from pyteal.types import TealType
from pyteal.ir import TealOp, Op, TealBlock

from pyteal.ast.leafexpr import LeafExpr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class MethodSignature(LeafExpr):
    """An expression that represents an ABI method selector"""

    def __init__(self, methodName: str) -> None:
        """Create a new method selector for ABI method call.

        Args:
            methodName: A string containing a valid ABI method signature
        """
        super().__init__()
        if type(methodName) is not str:
            raise TealInputError(
                "invalid input type {} to Method".format(type(methodName))
            )
        elif len(methodName) == 0:
            raise TealInputError("invalid input empty string to Method")
        self.methodName = methodName

    def __teal__(self, options: "CompileOptions"):
        op = TealOp(self, Op.method_signature, '"{}"'.format(self.methodName))
        return TealBlock.FromOp(options, op)

    def __str__(self) -> str:
        return "(MethodSignature '{}')".format(self.methodName)

    def type_of(self) -> TealType:
        return TealType.bytes


MethodSignature.__module__ = "pyteal"



import pytest

import pyteal as pt

avm2Options = pt.CompileOptions(version=2)
avm3Options = pt.CompileOptions(version=3)
avm5Options = pt.CompileOptions(version=5)
avm6Options = pt.CompileOptions(version=6)


def test_global_min_txn_fee():
    expr = pt.Global.min_txn_fee()
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.global_, "MinTxnFee")])

    actual, _ = expr.__teal__(avm2Options)

    assert actual == expected


def test_global_min_balance():
    expr = pt.Global.min_balance()
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.global_, "MinBalance")])

    actual, _ = expr.__teal__(avm2Options)

    assert actual == expected


def test_global_max_txn_life():
    expr = pt.Global.max_txn_life()
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.global_, "MaxTxnLife")])

    actual, _ = expr.__teal__(avm2Options)

    assert actual == expected


def test_global_zero_address():
    expr = pt.Global.zero_address()
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.global_, "ZeroAddress")])

    actual, _ = expr.__teal__(avm2Options)

    assert actual == expected


def test_global_group_size():
    expr = pt.Global.group_size()
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.global_, "GroupSize")])

    actual, _ = expr.__teal__(avm2Options)

    assert actual == expected


def test_global_logic_sig_version():
    expr = pt.Global.logic_sig_version()
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.global_, "LogicSigVersion")])

    actual, _ = expr.__teal__(avm2Options)

    assert actual == expected


def test_global_round():
    expr = pt.Global.round()
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.global_, "Round")])

    actual, _ = expr.__teal__(avm2Options)

    assert actual == expected


def test_global_latest_timestamp():
    expr = pt.Global.latest_timestamp()
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.global_, "LatestTimestamp")])

    actual, _ = expr.__teal__(avm2Options)

    assert actual == expected


def test_global_current_application_id():
    expr = pt.Global.current_application_id()
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [pt.TealOp(expr, pt.Op.global_, "CurrentApplicationID")]
    )

    actual, _ = expr.__teal__(avm2Options)

    assert actual == expected


def test_global_creator_address():
    expr = pt.Global.creator_address()
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.global_, "CreatorAddress")])

    actual, _ = expr.__teal__(avm3Options)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm2Options)


def test_global_current_application_address():
    expr = pt.Global.current_application_address()
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [pt.TealOp(expr, pt.Op.global_, "CurrentApplicationAddress")]
    )

    actual, _ = expr.__teal__(avm5Options)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_global_group_id():
    expr = pt.Global.group_id()
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.global_, "GroupID")])

    actual, _ = expr.__teal__(avm5Options)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_global_opcode_budget():
    expr = pt.Global.opcode_budget()
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.global_, "OpcodeBudget")])

    actual, _ = expr.__teal__(avm6Options)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm5Options)


def test_global_caller_application_id():
    expr = pt.Global.caller_app_id()
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [pt.TealOp(expr, pt.Op.global_, "CallerApplicationID")]
    )

    actual, _ = expr.__teal__(avm6Options)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm5Options)


def test_global_caller_app_address():
    expr = pt.Global.caller_app_address()
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [pt.TealOp(expr, pt.Op.global_, "CallerApplicationAddress")]
    )

    actual, _ = expr.__teal__(avm6Options)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm5Options)



from typing import cast, TYPE_CHECKING, Optional

from pyteal.types import TealType, require_type
from pyteal.config import NUM_SLOTS
from pyteal.errors import TealInputError, TealInternalError
from pyteal.ast.expr import Expr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class ScratchSlot:
    """Represents the allocation of a scratch space slot."""

    # Unique identifier for the compiler to automatically assign slots
    # The id field is used by the compiler to map to an actual slot in the source code
    # Slot ids under 256 are manually reserved slots
    nextSlotId = NUM_SLOTS

    def __init__(self, requestedSlotId: int = None):
        """Initializes a scratch slot with a particular id

        Args:
            requestedSlotId (optional): A scratch slot id that the compiler must store the value.
                This id may be a Python int in the range [0-256).
        """
        if requestedSlotId is None:
            self.id = ScratchSlot.nextSlotId
            ScratchSlot.nextSlotId += 1
            self.isReservedSlot = False
        else:
            if requestedSlotId < 0 or requestedSlotId >= NUM_SLOTS:
                raise TealInputError(
                    "Invalid slot ID {}, should be in [0, {})".format(
                        requestedSlotId, NUM_SLOTS
                    )
                )
            self.id = requestedSlotId
            self.isReservedSlot = True

    def store(self, value: Expr = None) -> Expr:
        """Get an expression to store a value in this slot.

        Args:
            value (optional): The value to store in this slot.
                If not included, the last value on the stack will be stored.
                NOTE: storing the last value on the stack breaks the typical
                semantics of PyTeal, only use if you know what you're doing.
        """
        if value is not None:
            return ScratchStore(self, value)
        return ScratchStackStore(self)

    def load(self, type: TealType = TealType.anytype) -> "ScratchLoad":
        """Get an expression to load a value from this slot.

        Args:
            type (optional): The type being loaded from this slot, if known. Defaults to
                TealType.anytype.
        """
        return ScratchLoad(self, type)

    def index(self) -> "ScratchIndex":
        return ScratchIndex(self)

    def __repr__(self):
        return "ScratchSlot({})".format(self.id)

    def __str__(self):
        return "slot#{}".format(self.id)


ScratchSlot.__module__ = "pyteal"


class ScratchIndex(Expr):
    def __init__(self, slot: ScratchSlot):
        super().__init__()
        self.slot = slot

    def __str__(self):
        return "(ScratchIndex {})".format(self.slot)

    def type_of(self):
        return TealType.uint64

    def has_return(self):
        return False

    def __teal__(self, options: "CompileOptions"):
        from pyteal.ir import TealOp, Op, TealBlock

        op = TealOp(self, Op.int, self.slot)
        return TealBlock.FromOp(options, op)


ScratchIndex.__module__ = "pyteal"


class ScratchLoad(Expr):
    """Expression to load a value from scratch space."""

    def __init__(
        self,
        slot: ScratchSlot = None,
        type: TealType = TealType.anytype,
        index_expression: Expr = None,
    ):
        """Create a new ScratchLoad expression.

        Args:
            slot (optional): The slot to load the value from.
            type (optional): The type being loaded from this slot, if known. Defaults to
                TealType.anytype.
            index_expression (optional): As an alternative to slot,
                an expression can be supplied for the slot index.
        """
        super().__init__()

        if (slot is None) == (index_expression is None):
            raise TealInputError(
                "Exactly one of slot or index_expressions must be provided"
            )

        if index_expression:
            if not isinstance(index_expression, Expr):
                raise TealInputError(
                    "index_expression must be an Expr but was of type {}".format(
                        type(index_expression)
                    )
                )
            require_type(index_expression, TealType.uint64)

        if slot and not isinstance(slot, ScratchSlot):
            raise TealInputError(
                "cannot handle slot of type {}".format(type(self.slot))
            )

        self.slot = slot
        self.type = type
        self.index_expression = index_expression

    def __str__(self):
        return "(Load {})".format(self.slot if self.slot else self.index_expression)

    def __teal__(self, options: "CompileOptions"):
        from pyteal.ir import TealOp, Op, TealBlock

        if self.index_expression is not None:
            op = TealOp(self, Op.loads)
            return TealBlock.FromOp(options, op, self.index_expression)

        s = cast(ScratchSlot, self.slot)
        op = TealOp(self, Op.load, s)
        return TealBlock.FromOp(options, op)

    def type_of(self):
        return self.type

    def has_return(self):
        return False


ScratchLoad.__module__ = "pyteal"


class ScratchStore(Expr):
    """Expression to store a value in scratch space."""

    def __init__(
        self, slot: Optional[ScratchSlot], value: Expr, index_expression: Expr = None
    ):
        """Create a new ScratchStore expression.

        Args:
            slot (optional): The slot to store the value in.
            value: The value to store.
            index_expression (optional): As an alternative to slot,
                an expression can be supplied for the slot index.
        """
        super().__init__()

        if (slot is None) == (index_expression is None):
            raise TealInternalError(
                "Exactly one of slot or index_expressions must be provided"
            )

        if index_expression:
            if not isinstance(index_expression, Expr):
                raise TealInputError(
                    "index_expression must be an Expr but was of type {}".format(
                        type(index_expression)
                    )
                )
            require_type(index_expression, TealType.uint64)

        self.slot = slot
        self.value = value
        self.index_expression = index_expression

    def __str__(self):
        return "(Store {} {})".format(
            self.slot if self.slot else self.index_expression, self.value
        )

    def __teal__(self, options: "CompileOptions"):
        from pyteal.ir import TealOp, Op, TealBlock

        if self.index_expression is not None:
            op = TealOp(self, Op.stores)
            return TealBlock.FromOp(options, op, self.index_expression, self.value)

        if not isinstance(self.slot, ScratchSlot):
            raise TealInternalError(
                "cannot handle slot of type {}".format(type(self.slot))
            )
        op = TealOp(self, Op.store, self.slot)
        return TealBlock.FromOp(options, op, self.value)

    def type_of(self):
        return TealType.none

    def has_return(self):
        return False


ScratchStore.__module__ = "pyteal"


class ScratchStackStore(Expr):
    """Expression to store a value from the stack in scratch space.

    NOTE: This expression breaks the typical semantics of PyTeal, only use if you know what you're
    doing.
    """

    def __init__(self, slot: ScratchSlot):
        """Create a new ScratchStackStore expression.

        Args:
            slot: The slot to store the value in.
        """
        super().__init__()
        self.slot = slot

    def __str__(self):
        return "(StackStore {})".format(self.slot)

    def __teal__(self, options: "CompileOptions"):
        from pyteal.ir import TealOp, Op, TealBlock

        op = TealOp(self, Op.store, self.slot)
        return TealBlock.FromOp(options, op)

    def type_of(self):
        return TealType.none

    def has_return(self):
        return False


ScratchStackStore.__module__ = "pyteal"



import pytest

import pyteal as pt

avm6Options = pt.CompileOptions(version=6)
avm7Options = pt.CompileOptions(version=7)


def test_base64decode_std():
    arg = pt.Bytes("aGVsbG8gd29ybGQ=")
    expr = pt.Base64Decode.std(arg)
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.byte, '"aGVsbG8gd29ybGQ="'),
            pt.TealOp(expr, pt.Op.base64_decode, "StdEncoding"),
        ]
    )

    actual, _ = expr.__teal__(avm7Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm6Options)


def test_base64decode_url():
    arg = pt.Bytes("aGVsbG8gd29ybGQ")
    expr = pt.Base64Decode.url(arg)
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.byte, '"aGVsbG8gd29ybGQ"'),
            pt.TealOp(expr, pt.Op.base64_decode, "URLEncoding"),
        ]
    )

    actual, _ = expr.__teal__(avm7Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm6Options)


def test_base64decode_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Base64Decode.std(pt.Int(0))
    with pytest.raises(pt.TealTypeError):
        pt.Base64Decode.url(pt.Int(0))



from typing import TYPE_CHECKING
from enum import Enum

from pyteal.types import TealType, require_type
from pyteal.errors import verifyFieldVersion
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.ast.expr import Expr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class Base64Encoding(Enum):
    # fmt: off
    #     id  |   name   | min version
    url = (0, "URLEncoding", 7)
    std = (1, "StdEncoding", 7)
    # fmt: on

    def __init__(self, id: int, name: str, min_version: int) -> None:
        self.id = id
        self.arg_name = name
        self.min_version = min_version


Base64Encoding.__module__ = "pyteal"


class Base64Decode(Expr):
    """An expression that decodes a base64-encoded byte string according to a specific encoding.

    See [RFC 4648](https://rfc-editor.org/rfc/rfc4648.html#section-4) (sections 4 and 5) for information on specifications.

    It is assumed that the encoding ends with the exact number of = padding characters as required by the RFC.
    When padding occurs, any unused pad bits in the encoding must be set to zero or the decoding will fail.
    The special cases of \\n and \\r are allowed but completely ignored. An error will result when attempting
    to decode a string with a character that is not in the encoding alphabet or not one of =, \\r, or \\n.

    NOTE:  Base64Decode usage is not intended for introducing constants. Instead, use :any:`Bytes`.
    """

    def __init__(self, encoding: Base64Encoding, base64: Expr) -> None:
        super().__init__()
        self.encoding = encoding

        require_type(base64, TealType.bytes)
        self.base64 = base64

    def __teal__(self, options: "CompileOptions"):
        verifyFieldVersion(
            self.encoding.arg_name, self.encoding.min_version, options.version
        )

        op = TealOp(self, Op.base64_decode, self.encoding.arg_name)
        return TealBlock.FromOp(options, op, self.base64)

    def __str__(self):
        return "(Base64Decode {})".format(self.encoding.arg_name)

    def type_of(self):
        return TealType.bytes

    def has_return(self):
        return False

    @classmethod
    def url(cls, base64: Expr) -> Expr:
        """Decode a base64-encoded byte string according to the URL encoding.

        Refer to the `Base64Decode` class documentation for more information.

        Args:
            base64: A base64-encoded byte string.
        """
        return cls(Base64Encoding.url, base64)

    @classmethod
    def std(cls, base64: Expr) -> Expr:
        """Decode a base64-encoded byte string according to the Standard encoding.

        Refer to the `Base64Decode` class documentation for more information.

        Args:
            base64: A base64-encoded byte string.
        """
        return cls(Base64Encoding.std, base64)


Base64Decode.__module__ = "pyteal"



import pytest
import pkg_resources
from tests.mock_version import (  # noqa: F401
    mock_version,
)

import pyteal as pt
from pyteal.pragma.pragma import __convert_pep440_compiler_version


@pytest.mark.parametrize(
    "compiler_version,expected",
    [
        ("2", "2.0.0"),
        (">=0.12.0a9.post2.dev9", ">=0.12.0-alpha9.2.9"),
        ("<0.5.0 || >=1.0.0a9.post1.dev2", "<0.5.0 || >=1.0.0-alpha9.1.2"),
        ("v0.12.9.post1 - v1.13.0.dev1", "0.12.9-1.0 - 1.13.0-0.1"),
        (
            "1.2.3a4.post5.dev6+AVM7.1",
            "1.2.3-alpha4.5.6+avm7.1",
        ),  # local versions are lowercased to be consistent with pkg_resources
    ],
)
def test_convert_pep440_compiler_version(compiler_version, expected):
    assert __convert_pep440_compiler_version(compiler_version) == expected


@pytest.mark.usefixtures("mock_version")
@pytest.mark.parametrize(
    "version, compiler_version, should_error",
    [
        # valid
        ("0.12.0", "0.12.0", False),
        ("0.12.0", "<=0.12.0", False),
        ("0.12.0", ">=0.12.0", False),
        ("0.13.0", "<0.8.0 || >=0.12.0", False),
        ("0.12.0", "0.12.0-rc1", False),
        ("0.1.0", "<0.2.0", False),
        ("1.2.3", "^1.2.3", False),
        ("1.5.0", "^1.2.3", False),
        ("1.2.3b9", "^1.2.3b4", False),
        ("0.1.0a1", "<0.1.0a2", False),
        ("0.1.0-rc1", "<0.1.0-rc2", False),
        ("0.1.0.dev1", "<0.1.0.dev2", False),
        ("0.1.0a9.dev2", ">0.1.0a8.dev1", False),
        ("v1.1", "<0.5.0 || >=1.0.0a9.post1.dev2", False),
        ("v1.0a9.post2.dev2", "<0.5.0 || >=1.0.0a9.post1.dev2", False),
        ("v1.0a9.post1", "<0.5.0 || >=1.0.0a9.dev10", False),
        ("0.4.0", "<0.5.0 || >=1.0.0a9.dev10", False),
        ("1.2.3a4.post5.dev6+AVM7.1", "=1.2.3a4.post5.dev6+AVM7.1", False),
        (
            "1.0.0+AVM7.1",
            "=1.0.0",
            False,
        ),  # Ignores local version (consistent with PEP 440)
        (
            pkg_resources.require("pyteal")[0].version,
            pkg_resources.require("pyteal")[0].version,
            False,
        ),
        # invalid
        ("0.13.0", "0.13.1", True),
        ("1.2.3a2", "<0.8.0 || >=0.12.0", True),
        ("0.1.0a1", "<0.2.0", True),
        ("2.0.0", "^1.2.3", True),
        ("0.4.0b10", "<0.5.0 || >=1.0.0a9.dev10", True),
        ("0.4.9a10.dev2.post3", "<0.5.0 || >=1.0.0a9.post1.dev2", True),
    ],
)
def test_pragma_compiler_version(version, compiler_version, should_error):
    if should_error:
        with pytest.raises(pt.TealPragmaError):
            pt.pragma(compiler_version=compiler_version)
    else:
        pt.pragma(compiler_version=compiler_version)


@pytest.mark.parametrize(
    "compiler_version",
    ["not a version", ">=0.1.1,<0.3.0", "1.2.3aq"],  # incorrect spec  # invalid PEP 440
)
def test_pragma_compiler_version_invalid(compiler_version):
    with pytest.raises(ValueError):
        pt.pragma(compiler_version=compiler_version)



from pyteal.compiler.optimizer.optimizer import OptimizeOptions, _apply_slot_to_stack

import pyteal as pt


def test_optimize_single_block():
    slot1 = pt.ScratchSlot(1)
    slot2 = pt.ScratchSlot(2)

    # empty check
    empty_block = pt.TealSimpleBlock([])
    _apply_slot_to_stack(empty_block, empty_block, {})

    expected = pt.TealSimpleBlock([])
    assert empty_block == expected

    # basic optimization
    block = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.store, slot1),
            pt.TealOp(None, pt.Op.load, slot1),
        ]
    )
    _apply_slot_to_stack(block, block, {})

    expected = pt.TealSimpleBlock([])
    assert block == expected

    # iterate optimization
    block = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.store, slot1),
            pt.TealOp(None, pt.Op.store, slot2),
            pt.TealOp(None, pt.Op.load, slot2),
            pt.TealOp(None, pt.Op.load, slot1),
        ]
    )
    _apply_slot_to_stack(block, block, {})

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.store, slot1),
            pt.TealOp(None, pt.Op.load, slot1),
        ]
    )
    assert block == expected

    _apply_slot_to_stack(block, block, {})
    expected = pt.TealSimpleBlock([])
    assert block == expected

    # remove extraneous stores
    block = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.store, slot1),
            pt.TealOp(None, pt.Op.load, slot1),
            pt.TealOp(None, pt.Op.store, slot1),
        ]
    )
    _apply_slot_to_stack(block, block, {})

    expected = pt.TealSimpleBlock([])
    assert block == expected


def test_optimize_subroutine():
    @pt.Subroutine(pt.TealType.uint64)
    def add(a1: pt.Expr, a2: pt.Expr) -> pt.Expr:
        return a1 + a2

    program = pt.Seq(
        [
            pt.If(pt.Txn.sender() == pt.Global.creator_address()).Then(
                pt.Pop(add(pt.Int(1), pt.Int(2)))
            ),
            pt.Approve(),
        ]
    )

    optimize_options = OptimizeOptions()

    # unoptimized
    expected = """#pragma version 4
txn Sender
global CreatorAddress
==
bz main_l2
int 1
int 2
callsub add_0
pop
main_l2:
int 1
return

// add
add_0:
store 1
store 0
load 0
load 1
+
retsub
    """.strip()
    actual = pt.compileTeal(
        program, version=4, mode=pt.Mode.Application, optimize=optimize_options
    )
    assert actual == expected

    # optimized
    expected = """#pragma version 4
txn Sender
global CreatorAddress
==
bz main_l2
int 1
int 2
callsub add_0
pop
main_l2:
int 1
return

// add
add_0:
+
retsub
    """.strip()
    optimize_options = OptimizeOptions(scratch_slots=True)
    actual = pt.compileTeal(
        program, version=4, mode=pt.Mode.Application, optimize=optimize_options
    )
    assert actual == expected


def test_optimize_subroutine_with_scratchvar_arg():
    @pt.Subroutine(pt.TealType.uint64)
    def add(a1: pt.ScratchVar, a2: pt.Expr) -> pt.Expr:
        return a1.load() + a2

    arg = pt.ScratchVar(pt.TealType.uint64)

    program = pt.Seq(
        [
            arg.store(pt.Int(1)),
            pt.If(pt.Txn.sender() == pt.Global.creator_address()).Then(
                pt.Pop(add(arg, pt.Int(2)))
            ),
            pt.Approve(),
        ]
    )

    optimize_options = OptimizeOptions()

    # unoptimized
    expected = """#pragma version 5
int 1
store 0
txn Sender
global CreatorAddress
==
bz main_l2
int 0
int 2
callsub add_0
pop
main_l2:
int 1
return

// add
add_0:
store 2
store 1
load 1
loads
load 2
+
retsub""".strip()
    actual = pt.compileTeal(
        program, version=5, mode=pt.Mode.Application, optimize=optimize_options
    )
    assert actual == expected

    # optimized
    expected = """#pragma version 5
int 1
store 0
txn Sender
global CreatorAddress
==
bz main_l2
int 0
int 2
callsub add_0
pop
main_l2:
int 1
return

// add
add_0:
store 1
loads
load 1
+
retsub""".strip()
    optimize_options = OptimizeOptions(scratch_slots=True)
    actual = pt.compileTeal(
        program, version=5, mode=pt.Mode.Application, optimize=optimize_options
    )
    assert actual == expected


def test_optimize_subroutine_with_local_var():
    local_var = pt.ScratchVar(pt.TealType.uint64)

    @pt.Subroutine(pt.TealType.uint64)
    def add(a1: pt.Expr) -> pt.Expr:
        return pt.Seq(local_var.store(pt.Int(2)), local_var.load() + a1)

    program = pt.Seq(
        [
            pt.If(pt.Txn.sender() == pt.Global.creator_address()).Then(
                pt.Pop(add(pt.Int(1)))
            ),
            pt.Approve(),
        ]
    )

    optimize_options = OptimizeOptions()

    # unoptimized
    expected = """#pragma version 4
txn Sender
global CreatorAddress
==
bz main_l2
int 1
callsub add_0
pop
main_l2:
int 1
return

// add
add_0:
store 1
int 2
store 0
load 0
load 1
+
retsub
    """.strip()
    actual = pt.compileTeal(
        program, version=4, mode=pt.Mode.Application, optimize=optimize_options
    )
    assert actual == expected

    # optimized
    expected = """#pragma version 4
txn Sender
global CreatorAddress
==
bz main_l2
int 1
callsub add_0
pop
main_l2:
int 1
return

// add
add_0:
store 0
int 2
load 0
+
retsub
    """.strip()
    optimize_options = OptimizeOptions(scratch_slots=True)
    actual = pt.compileTeal(
        program, version=4, mode=pt.Mode.Application, optimize=optimize_options
    )
    assert actual == expected


def test_optimize_subroutine_with_global_var():
    global_var = pt.ScratchVar(pt.TealType.uint64)

    @pt.Subroutine(pt.TealType.uint64)
    def add(a1: pt.Expr) -> pt.Expr:
        return pt.Seq(global_var.store(pt.Int(2)), global_var.load() + a1)

    program = pt.Seq(
        [
            pt.If(pt.Txn.sender() == pt.Global.creator_address()).Then(
                pt.Pop(add(pt.Int(1)))
            ),
            global_var.store(pt.Int(5)),
            pt.Approve(),
        ]
    )

    optimize_options = OptimizeOptions()

    # unoptimized
    expected = """#pragma version 4
txn Sender
global CreatorAddress
==
bz main_l2
int 1
callsub add_0
pop
main_l2:
int 5
store 0
int 1
return

// add
add_0:
store 1
int 2
store 0
load 0
load 1
+
retsub
    """.strip()
    actual = pt.compileTeal(
        program, version=4, mode=pt.Mode.Application, optimize=optimize_options
    )
    assert actual == expected

    # optimization should not apply to global vars
    optimize_options = OptimizeOptions(scratch_slots=True)
    actual = pt.compileTeal(
        program, version=4, mode=pt.Mode.Application, optimize=optimize_options
    )
    assert actual == expected


def test_optimize_subroutine_with_reserved_local_var():
    local_var = pt.ScratchVar(pt.TealType.uint64, 0)

    @pt.Subroutine(pt.TealType.uint64)
    def add(a1: pt.Expr) -> pt.Expr:
        return pt.Seq(local_var.store(pt.Int(2)), local_var.load() + a1)

    program = pt.Seq(
        [
            pt.If(pt.Txn.sender() == pt.Global.creator_address()).Then(
                pt.Pop(add(pt.Int(1)))
            ),
            pt.Approve(),
        ]
    )

    optimize_options = OptimizeOptions()

    # unoptimized
    expected = """#pragma version 4
txn Sender
global CreatorAddress
==
bz main_l2
int 1
callsub add_0
pop
main_l2:
int 1
return

// add
add_0:
store 1
int 2
store 0
load 0
load 1
+
retsub
    """.strip()
    actual = pt.compileTeal(
        program, version=4, mode=pt.Mode.Application, optimize=optimize_options
    )
    assert actual == expected

    # The optimization must skip over the reserved slot id so the expected result
    # hasn't changed.
    optimize_options = OptimizeOptions(scratch_slots=True)
    actual = pt.compileTeal(
        program, version=4, mode=pt.Mode.Application, optimize=optimize_options
    )
    assert actual == expected


def test_optimize_subroutine_with_load_dependency():
    @pt.Subroutine(pt.TealType.uint64)
    def add(a1: pt.Expr, a2: pt.Expr) -> pt.Expr:
        return pt.Seq(pt.Pop(a1 + a2), a2)

    program = pt.Seq(
        [
            pt.If(pt.Txn.sender() == pt.Global.creator_address()).Then(
                pt.Pop(add(pt.Int(1), pt.Int(2)))
            ),
            pt.Approve(),
        ]
    )

    optimize_options = OptimizeOptions()

    # unoptimized
    expected = """#pragma version 4
txn Sender
global CreatorAddress
==
bz main_l2
int 1
int 2
callsub add_0
pop
main_l2:
int 1
return

// add
add_0:
store 1
store 0
load 0
load 1
+
pop
load 1
retsub""".strip()
    actual = pt.compileTeal(
        program, version=4, mode=pt.Mode.Application, optimize=optimize_options
    )
    assert actual == expected

    # slot 0 will get optimized but slot 1 won't because it has
    # a load dependency.
    expected = """#pragma version 4
txn Sender
global CreatorAddress
==
bz main_l2
int 1
int 2
callsub add_0
pop
main_l2:
int 1
return

// add
add_0:
store 0
load 0
+
pop
load 0
retsub""".strip()
    optimize_options = OptimizeOptions(scratch_slots=True)
    actual = pt.compileTeal(
        program, version=4, mode=pt.Mode.Application, optimize=optimize_options
    )
    assert actual == expected


def test_optimize_multi_value():
    # note: this is incorrect usage of the app_global_get_ex opcode
    program = pt.Seq(
        pt.MultiValue(
            pt.Op.app_global_get_ex,
            [pt.TealType.uint64, pt.TealType.uint64],
            immediate_args=[],
            args=[pt.Int(0), pt.Int(1)],
        ).outputReducer(lambda value, hasValue: pt.Pop(value + hasValue)),
        pt.Approve(),
    )

    optimize_options = OptimizeOptions()

    # unoptimized
    expected = """#pragma version 4
int 0
int 1
app_global_get_ex
store 1
store 0
load 0
load 1
+
pop
int 1
return""".strip()
    actual = pt.compileTeal(
        program, version=4, mode=pt.Mode.Application, optimize=optimize_options
    )
    assert actual == expected

    # optimized
    expected = """#pragma version 4
int 0
int 1
app_global_get_ex
+
pop
int 1
return""".strip()
    optimize_options = OptimizeOptions(scratch_slots=True)
    actual = pt.compileTeal(
        program, version=4, mode=pt.Mode.Application, optimize=optimize_options
    )
    assert actual == expected


def test_optimize_dynamic_var():
    myvar = pt.DynamicScratchVar()
    regvar = pt.ScratchVar()
    program = pt.Seq(
        regvar.store(pt.Int(1)),
        myvar.set_index(regvar),
        regvar.store(pt.Int(2)),
        pt.Pop(regvar.load()),
        pt.Approve(),
    )

    optimize_options = OptimizeOptions()

    # unoptimized
    expected = """#pragma version 4
int 1
store 1
int 1
store 0
int 2
store 1
load 1
pop
int 1
return""".strip()
    actual = pt.compileTeal(
        program, version=4, mode=pt.Mode.Application, optimize=optimize_options
    )
    assert actual == expected

    # optimization should not change the code because the candidate slot
    # is used by the dynamic slot variable.
    optimize_options = OptimizeOptions(scratch_slots=True)
    actual = pt.compileTeal(
        program, version=4, mode=pt.Mode.Application, optimize=optimize_options
    )
    assert actual == expected



from typing import List, Tuple, TYPE_CHECKING

from pyteal.types import TealType
from pyteal.errors import TealInternalError, TealCompileError
from pyteal.ir import TealOp, Op, TealSimpleBlock
from pyteal.ast.expr import Expr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


def multiplyFactors(
    expr: Expr, factors: List[Expr], options: "CompileOptions"
) -> Tuple[TealSimpleBlock, TealSimpleBlock]:
    if len(factors) == 0:
        raise TealInternalError("Received 0 factors")

    start = TealSimpleBlock([])

    fac0Start, fac0End = factors[0].__teal__(options)

    if len(factors) == 1:
        # need to use 0 as high word
        highword = TealSimpleBlock([TealOp(expr, Op.int, 0)])

        start.setNextBlock(highword)
        highword.setNextBlock(fac0Start)

        end = fac0End
    else:
        start.setNextBlock(fac0Start)

        fac1Start, fac1End = factors[1].__teal__(options)
        fac0End.setNextBlock(fac1Start)

        multiplyFirst2 = TealSimpleBlock([TealOp(expr, Op.mulw)])
        fac1End.setNextBlock(multiplyFirst2)

        end = multiplyFirst2
        for factor in factors[2:]:
            facXStart, facXEnd = factor.__teal__(options)
            end.setNextBlock(facXStart)

            # stack is [..., A, B, C], where C is current factor
            # need to pop all A,B,C from stack and push X,Y, where X and Y are:
            #       X * 2**64 + Y = (A * 2**64 + B) * C
            # <=>   X * 2**64 + Y = A * C * 2**64 + B * C
            # <=>   X = A * C + highword(B * C)
            #       Y = lowword(B * C)
            multiply = TealSimpleBlock(
                [
                    TealOp(expr, Op.uncover, 2),  # stack: [..., B, C, A]
                    TealOp(expr, Op.dig, 1),  # stack: [..., B, C, A, C]
                    TealOp(expr, Op.mul),  # stack: [..., B, C, A*C]
                    TealOp(expr, Op.cover, 2),  # stack: [..., A*C, B, C]
                    TealOp(
                        expr, Op.mulw
                    ),  # stack: [..., A*C, highword(B*C), lowword(B*C)]
                    TealOp(
                        expr, Op.cover, 2
                    ),  # stack: [..., lowword(B*C), A*C, highword(B*C)]
                    TealOp(
                        expr, Op.add
                    ),  # stack: [..., lowword(B*C), A*C+highword(B*C)]
                    TealOp(
                        expr, Op.swap
                    ),  # stack: [..., A*C+highword(B*C), lowword(B*C)]
                ]
            )

            facXEnd.setNextBlock(multiply)
            end = multiply

    return start, end


class WideRatio(Expr):
    """A class used to calculate expressions of the form :code:`(N_1 * N_2 * N_3 * ...) / (D_1 * D_2 * D_3 * ...)`

    Use this class if all inputs to the expression are uint64s, the output fits in a uint64, and all
    intermediate values fit in a uint128.
    """

    def __init__(
        self, numeratorFactors: List[Expr], denominatorFactors: List[Expr]
    ) -> None:
        """Create a new WideRatio expression with the given numerator and denominator factors.

        This will calculate :code:`(N_1 * N_2 * N_3 * ...) / (D_1 * D_2 * D_3 * ...)`, where each
        :code:`N_i` represents an element in :code:`numeratorFactors` and each :code:`D_i`
        represents an element in :code:`denominatorFactors`.

        Requires program version 5 or higher.

        Args:
            numeratorFactors: The factors in the numerator of the ratio. This list must have at
                least 1 element. If this list has exactly 1 element, then denominatorFactors must
                have more than 1 element (otherwise basic division should be used).
            denominatorFactors: The factors in the denominator of the ratio. This list must have at
                least 1 element.
        """
        super().__init__()
        if len(numeratorFactors) == 0 or len(denominatorFactors) == 0:
            raise TealInternalError(
                "At least 1 factor must be present in the numerator and denominator"
            )
        if len(numeratorFactors) == 1 and len(denominatorFactors) == 1:
            raise TealInternalError(
                "There is only a single factor in the numerator and denominator. Use basic division instead."
            )
        self.numeratorFactors = numeratorFactors
        self.denominatorFactors = denominatorFactors

    def __teal__(self, options: "CompileOptions"):
        if options.version < Op.cover.min_version:
            raise TealCompileError(
                "WideRatio requires program version {} or higher".format(
                    Op.cover.min_version
                ),
                self,
            )

        numStart, numEnd = multiplyFactors(self, self.numeratorFactors, options)
        denomStart, denomEnd = multiplyFactors(self, self.denominatorFactors, options)
        numEnd.setNextBlock(denomStart)

        combine = TealSimpleBlock(
            [
                TealOp(self, Op.divmodw),
                TealOp(self, Op.pop),  # pop remainder low word
                TealOp(self, Op.pop),  # pop remainder high word
                TealOp(self, Op.swap),  # swap quotient high and low words
                TealOp(self, Op.logic_not),
                TealOp(self, Op.assert_),  # assert quotient high word is 0
                # end with quotient low word remaining on the stack
            ]
        )
        denomEnd.setNextBlock(combine)

        return numStart, combine

    def __str__(self):
        ret_str = "(WideRatio (*"
        for f in self.numeratorFactors:
            ret_str += " " + str(f)
        ret_str += ") (*"
        for f in self.denominatorFactors:
            ret_str += " " + str(f)
        ret_str += ")"
        return ret_str

    def type_of(self):
        return TealType.uint64

    def has_return(self):
        return False


WideRatio.__module__ = "pyteal"
